/* ********************************************************************
* Interleave
* Copyright (c) 2001-2016 Hidde Fennema (info@interleave.nl)
* Licensed under the GNU GPL. For full terms see http://www.gnu.org/
*
* This script handles all form-related javascript for records & lists
*
* Check http://www.interleave.nl/ for more information
* All functions in this script must be called with the fully qualified field name + the recordid
*
* BUILD / RELEASE :: 6.1.0.20150508
*
**********************************************************************
*/

/*

function general_form_submit

Submits any form after validating it. This functions gathers all input
elements in the form and creates a JSON array of it. This array is posted
to the webserver for validation. If all fields pass, the form is submmited.
If some don't, they are highlighted in the form and submitting is canceled.

var formid						the id of the form (or div) being submitted (typically t12-23024 or t45-_new_)
var recordid					the concerning record
var save_after_validate			tell server to treat this as a save, next to validate
*/

function general_form_submit(formid, recordid, save_after_validate, id_to_place_wait_on) {
	var msg = '';
	if (id_to_place_wait_on == '')
	{
		id_to_place_wait_on = formid;
	}


	tell_server_to_save_after_validation = 0;
	if (save_after_validate)
	{
		tell_server_to_save_after_validation = 1;
	}

	/* show wait image */
	SWID(id_to_place_wait_on);

	/* determine table */
	var parts = formid.split('-');
	var table = parts[0];
	var data_to_post = new Array();
	var elementcounter = 0;

	/* workaround; files are submitted from a div, not a form, and is has a slightly different id convention */
	if (table == 'JS_uploadfiles') {
		table = 'files';
	}

	/* we send the whole form to the server as a JSON array */
	$('#' + formid +" :input[name^='" + table + "-']").each(function () {

		/* if the field is disabled, ignore at, as will the POST to the server */
		if ($(this).is(':disabled')) {

			/*
			 *alert('not posting ' + $(this).attr('name') + ' not enabled');
			 **/

		} else {

			/* this is a field for my table */
			if ($(this).attr('type') == 'checkbox') {

				/* this is a checkbox. only add the value if checked, but add empty if unchecked to make sure the server knows it's posted */
				if ($(this).prop('checked')) {
					var localvalue = $(this).val();
				} else {
					var localvalue = '';
				}

			} else {

				/* not a checkbox. just add the value */
				var localvalue = $(this).val();
			}

			/* we don't post buttons */
			if ($(this).attr('type') != 'button') {

				/* create new array, add to data to post */
				var localarray = new Array(this.name, encodeURIComponent(localvalue));
				data_to_post.push(localarray);
				elementcounter++;
			}
		}
	});
	var all_ok = true;

	if (elementcounter > 0) {

		/* compose data string */

		var data = 'method=general_form_submit&fqn=' + formid + '&recordid=' + recordid + '&save_after_validation=' + tell_server_to_save_after_validation + '&value=' + encodeURIComponent(JSON.stringify(data_to_post));

		/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
		x = submit_call_to_server(data, '', 'sync');

		/* get results from array */
		method		= x[0];
		element		= x[1];
		recordid	= x[2];
		result		= x[3];

		if (typeof result["data"] == 'object') {
			for (var i = 0; i < result["data"].length ; i++) {

				returned_element = result["data"][i][0];
				checkresult = result["data"][i][1]["result"];

				if (checkresult == "nok") {

					/* something is wrong. place error marker. */
					if ($('#JS_' + returned_element + '-multiselect').length) {
						set_error(returned_element + '-multiselect');
					} else {
						set_error(returned_element);
					}

					all_ok = false;
					$("#JS_e_button").val('');

				} else {

					/* remove error marker. */
					if ($('#JS_' + returned_element + '-multiselect').length) {
						remove_error(returned_element + '-multiselect');
					} else {
						remove_error(returned_element);
					}

				}
			}

			/* and display a message */
			if (result["message"] != null && result["message"].length > 0) {

				/* it looks better if the wait-image-div is already gone */
				HWID(id_to_place_wait_on);

				/* the server js-encodes the message before sending it */
				alert(result["message"]);

			}
			if (result["directlysubmitted"] == 1)
			{
				/* this means the form was directly saved when validating (for progressive web app use only */

				/* exception for webapp; if we're saving directly after validating, we should, in case of a new record, redirect to the record using ILOR */
				if (result['openlink'] != '' && all_ok) {

					ilor(result['openlink']);
				}
				/* we return false to the caller to make sure the form doesn't get submitted */
				return(false);
			}

		}

		/* enable all input elements. disabled elements will not be posted so this is important! */
		if (all_ok === false) {
			/* HF 20161102 removed re-enebling form elements */
		} else {
			$('#JS_form_submitted').val('1');
		}
	} else {
		/*
		alert('Error: no elements found to submit.');
		*/
		all_ok = true;
	}


	return all_ok;

}

function update_diary(fqn, recordid, value, commenthash, delentry) {
	/*
		adds a value to a diary field, updates an entry, or deletes it.
	*/

	// alert('fqn ' + fqn + ' rec ' + recordid + ' val ' + value + ' commenthash ' + commenthash + ' delentry ' + delentry);

	/* assemble data to post to server */
	var data = 'method=update_diary&fqn=' + fqn + '&recordid=' + recordid + '&value=' + encodeURIComponent(value) + '&commenthash=' + commenthash + '&delentry=' + delentry;

	/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
	x = submit_call_to_server(data, '', 'sync');

	/* sync call, so we know all, except result */
	result		= x[3];

	/* check for messages */
	if (typeof result["message"] != 'undefined') {
		alert(result["message"]);
	}

	/* put the resulting table in the enclosing div */
	$('#diarycontent_' + fqn).html(result["result"]);

	/* empty the diary input field, if it's equal to value (to avoid emptying it on a refresh */
	if ($('#JS_' + fqn).val() == value && value != '')	{
		$('#JS_' + fqn).val('');
	}
	if (commenthash != '') {
		$('#JS_commenthash-' + fqn).val('');
	}


}


function reload_calculations(fqn, recordid, skip_lists) {
	/*
		This function reloads all calculations found in a page
	*/


	var data = 'method=reload_calculations&fqn=' + fqn + '&recordid=' + recordid;

	var elements = fqn.split('-');
	var table = elements[0];
	var field = elements[1];

	/* make sure we only return the values which are actually in the form */
	$('.calculation').each(function() {

		/* id syntax: listid--t12-f23--recordid - when on a form, the listid is always "" */
		test = $(this).prop("id").split("--");
		testlistid = test[0].substr(test[0].length - 7);

		var skip = false;
		if (skip_lists == true && testlistid != '') {
			skip = true;
		}

		if (test.length == 3 && skip == false)
		{

			var elements = test[1].split('-');
			var localtable = elements[0];
			var localfield = elements[1];

			/* check if it's for the current record and the same table OR a completely different table */
			if ((test[2] == recordid && table == localtable) || table != localtable)
			{
				data += '&value[]=' + $(this).prop("id");

			} else {

				/*

				NOT an error; jquery will cycle through all matching id's, only one row actually needs recalculation

				alert('error: call to reload a calculation of a foreign record id (e.g. not me: ' + test[2] + ' is not ' + recordid + '');


				*/
			}

		} else {
			/* help! */
			/* alert('error: unknown calculation id syntax:' + $(this).prop("id")); */

			/* ignore */
		}
	});

	/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
	/* alert(data); */
	x = submit_call_to_server(data, '', 'async');

}
function reload_calculations_callback(data, message) {
	/*
		data is like:
		row1 = [0] = element, [1] = new value
		row2 = [0] = element, [1] = new value

	*/

	for (i=0; i<data.length; i++) {

		/* place result in target in element */
		$('#' + data[i][0]).html(data[i][1]);

		/* place unparsed value in UPV */
		$('#' + data[i][0] + '-UPV').html(data[i][2]);

		/*
		 *alert('Placing resonse \'' + data[i][1] + '\' in element ' + data[1][0]);
		 */

		elements_of_id = data[i][0].split('--');

		/*
			Syntax of id

			calculations on a form: --FQN--RECORD
			calculations in a list: LISTID--FQN--RECORD

		*/
		if (elements_of_id[0] != '')
		{
			/* check if the resulting element has a span class colorspan */
			var bgchild = $('#' + data[i][0] + ' .colorspan').css('backgroundColor');
			var fgchild = $('#' + data[i][0] + ' .colorspan').css('color');

			if (typeof bgchild != 'undefined' && typeof fgchild != 'undefined')
			{
				$('#' + data[i][0]).css('background-color', bgchild);
				$('#' + data[i][0]).css('color', fgchild);
			} else {
			//	$('#' + data[i][0]).css('background-color', "inherit");
			//	$('#' + data[i][0]).css('color', "inherit");
			}
		}
		//$('#' + data[i][0]).css('background-color', 'green');
	}

	if (typeof message != 'undefined' && message != '')
	{
		alert(message);
	}

	/* update all clientside fields having client side validations */

	if (typeof hideshowallelements === 'function') {
		hideshowallelements();
	}



}

/*
function reload_element

Reloads any element on a form

*/

function reload_element(element, recordid, result, enclosure, ifield) {

	var value = '';

	var data = '';
	var some_data = false;

	/* check if var element is an array. if so, it contains multiple fqn's */
	if (Array.isArray(element)) {

		data += 'method=reload_element&recordid=' + recordid;

		/* cycle through array */
		element.forEach(function(Mfqn) {

			/* check if the element exists */
			if ($('#JS_' + Mfqn + '-' + recordid).length) {

				/* get the currently selected value */
				value = $('#JS_' + Mfqn + '-' + recordid).val();
				if (typeof value == 'undefined') {
					value = '';
				}

				/* assemble data to post to server */
				data += '&fqn[' + Mfqn + '][]=' + encodeURIComponent(value);
				data += '&fqn[' + Mfqn + '][]=' + encodeURIComponent(recordid);
				some_data = true;
			}
		});

	} else {
		alert('Passed reload set is not an array');
	}

	/* if we're reloading an ifield, we want the UNCLICKED state back. record-ajax.php returns the result of return_ifield if set */
	if (ifield == true) {
		data += '&ifield=1';
	}
	/* add all field values, reloads need them */
	$('[id^="JS_"]').each(function() {

		/* exclude buttons */
		if ($(this).attr('type') != 'button') {
			/* add value to array-to-submit */
			data += '&' + $(this).attr('name') + '=' + encodeURIComponent($(this).val());
		}
	});

	if (some_data) {
		/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
		x = submit_call_to_server(data, '', 'sync');
		method		= x[0];
		element		= x[1];
		recordid	= x[2];
		result		= x[3];


		$.each(result['result'], function (element, value) {

			enclosure = 'ffc-' + element;

			$('#' + enclosure).html(value);

			//$('#' + enclosure).css('background-color', 'blue');

			/* remove any error markers. is this safe? */
			remove_error(element + '-' + recordid);

		});

		if (typeof result["message"] != 'undefined' && result["message"] != '')	{
			/* the server jsencodes this */
			alert(result["message"]);
		}

		InitDatePicker();
	} else {
	/*	 alert('skipping reload, no data');  */
	}

}

function general_ifield_change(fqn, recordid, value, basetablerecordid, parentlistid, reload_list, result) {

	if (typeof result == 'undefined') {

		if (reload_list !== 0)
		{
			reload_list = true;
		}

		if (reload_list && recordid > 0 && parentlistid != '') {
			/* show wait image div */
			//SWID('MainContent');
		}
		/* get the text value */
		var textvalue = $( '#JS_' + fqn + ' option:selected' ).text();

		/* disable the element while loading */
		$('#JS_' + fqn ).prop('readonly', true);

		/* now check if this is a multiselect element */
		if ($('#JS_' + fqn + '-0-multiselect').length) {

			/* walk through the options add create an array of it */
			var vals = [];

			/* the div in which multiselects reside, walk through all checked checkboxes */
			var i = 0;
			$('#JS_' + fqn + '-multiselect').find(':input:checked').each(function() {

				/* add value to array-to-submit */
				vals[i] = $(this).val();
				i++;
			});

			/* create a JSONified version of the array */
			var val = JSON.stringify(vals);

		} else {
			/* it's a normal field, or at least we need the plain value */
			var val = $('#JS_' + fqn).val();
		}


		/* assemble data to post to server */
		var data = 'method=general_element_change&by=general_ifield_change&type=interactive&fqn=' + fqn + '&recordid=' + recordid + '&value=' + encodeURIComponent(val) + '&listid=' + parentlistid + '&reload_list=' + reload_list;

		/* submit call to server */
		//alert(data);
		var method = 'sync';
		if (reload_list != true)
		{
		/*
		Dit moet nog; maar daarvoor moet de hele structuur van deze functie omgegooid worden
		*/
			method = 'async';

		}

		$('[id^="JS_"]').each(function() {

			if ($(this).attr('name') != "recordid" && $(this).attr('name') != "fqn" && $(this).attr('name') != "value" && typeof $(this).attr('name') !== 'undefined' )
			{
				/* add value to array-to-submit */
				data += '&' + $(this).attr('name') + '=' + encodeURIComponent($(this).val());
			}
		});


		var x = submit_call_to_server(data, 'general_ifield_change', method);

	} else {

		var x = result;


	}

	if (typeof x != 'undefined' && x[0] == 'general_element_change') {

		method		= x[0];
		fqn			= x[1];
		recordid	= x[2];
		result		= x[3];

		if (result["result"] == "ok") {

			if (textvalue == '')
			{
				textvalue = nl2br(result["corrected_value"]);
			}
			if (textvalue == '')
			{
				/* this is because if we don't put something in here, the ifield underline will not be shown and cannot be clicked */
				textvalue = '&nbsp;&nbsp;&nbsp;';

			} else {
				/*
				alert(textvalue);
				*/
			}

			var elements = fqn.split('-');
			var fqn_to_submit = elements[0] + '-' + elements[1];
			var table = elements[0];
			var field = elements[1];
			var parentlistid = result['listid'];

			textvalue = '<span class="interactive_cell" id="interactive_cell_' + fqn + '" onclick="load_ifield(\'' + fqn_to_submit + '\',\'' + recordid + '\',\'' + parentlistid + '\');">' + textvalue + '</a>';

			/* remove any error classes if there */
			$('#JS_' + fqn ).removeClass('errorinfield');

			/* reload parent list, when given */
			if (typeof parentlistid != 'undefined' && parentlistid != '') {

				if (reload_list) {

					/* reload the list */
					/* tmp: issue met a65: als onderstaande uitgehast wordt srpingt de cursor niet naar boven */
					ListReload(parentlistid, true, true);

					/* reload calculations */
					reload_calculations(table + '-timestamp_last_change', recordid);

					/* very dirty but refreshing the list on all changes is deprecated anyhow */
					setTimeout(function(){

						/* move focus to next input element WITHIN THE LIST DIV */
						$('#' + parentlistid + ':input:eq(' + ($(':input').index($('#JS_' + fqn )) + 1) + ')').focus();

						/* select all contents of that field for easy editing */
						$('#' + parentlistid + ':input:eq(' + ($(':input').index($('#JS_' + fqn )) + 1) + ')').select();


					 }, 250); /* 250ms is very arbitrairy */

				} else {


					//HWID('MainContent');
					//alert('reload calcs ' + table + '-timestamp_last_change');

					/* reload calculations & computations */
					reload_calculations(table + '-timestamp_last_change', recordid);

					if (reload_list)
					{

						/* move focus to next input element WITHIN THE LIST DIV */
						$('#' + parentlistid + ':input:eq(' + ($(':input').index($('#JS_' + fqn )) + 1) + ')').focus();

						/* select all contents of that field for easy editing */
						$('#' + parentlistid + ':input:eq(' + ($(':input').index($('#JS_' + fqn )) + 1) + ')').select();
					}

					/* re-enable the element */
					$('#JS_' + fqn ).prop('readonly', false);

				}

			} else {

				/* update the ifield */
				$('#ifield-' + fqn).html(textvalue);

				/* re-enable the element */
				$('#JS_' + fqn ).prop('readonly', false);

				/* hide the wait-image div */
				HWID('MainContent');
			}


		} else {

			/* input didn't validate. we're not refreshing the list here, remove the wait image div */
			HWID('MainContent');

			/* re-enable the element */
			$('#JS_' + fqn ).prop('readonly', false);

			/* set error marker */
			$('#JS_' + fqn ).addClass('errorinfield');

			/* focus */
			$('#JS_' + fqn ).focus();

			/* select */
			$('#JS_' + fqn ).select();

			/* display generic message */
			/*
			alert(forcedfieldstext);
			*/

		}
	}

}
/*
function load_ifield

Loads an ifield

*/
function load_ifield(fqn, recordid, listid) {

	/* assemble data to post to server */

	var data = 'method=load_ifield&fqn=' + fqn + '&recordid=' + recordid + '&listid=' + listid;

	/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
	x = submit_call_to_server(data, '', 'sync');

	result		= x[3];

	/* put the result in the inner html of the div */
	$('#ifield-' + fqn + '-' + recordid).html(result["result"]);

	/* if there's an input element in the result, focus it */
	$('#ifield-' + fqn + '-' + recordid + ' input, #ifield-' + fqn + '-' + recordid + ' textarea').focus();
/*
$('#ifield-' + fqn + '-' + recordid + ' input, #ifield-' + fqn + '-' + recordid + ' textarea').select();
*/

	/* run init date picker in case it was a date type event */
	InitDatePicker();

	/* initialize any autosizing elements */
	$('.autosizeoninput').autosizeInput({
		space: 20
	});

	/* make sure enter doesn't submit the form */
	$('#JS_' + fqn ).keypress(function(event) {
		if (event.keyCode == 13)
		{
			$(this).blur();
		}
		return event.keyCode != 13;
	});

}

/*

function general_element_validate

Only validates input, does nothing else. Returns true|false.

Currently not in use.

*/
function general_element_validate(element, recordid, result) {

	/* assemble data to post to server */
	var data = 'method=general_element_validate&fqn=' + element + '&recordid=' + recordid + '&value=' + encodeURIComponent($('#JS_' + element).val());

		/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
	x = submit_call_to_server(data, '', 'sync');
	method		= x[0];
	element		= x[1];
	recordid	= x[2];
	result		= x[3];

	if (result["result"] == "ok") {

		/* extra-hightlight the field while showing the message */
		remove_error(element);

		return(true);
	}
	return false;
}


/*
function general_element_change

Is called on every field change (onchange). Handles validation, autosave

var element			the element name of the updated field, fully qualified ("t12-f34")
var recordid		the concerning recordid
var result			the result of the ajax call, in case of callback
*/
function general_element_change(element, recordid, result) {

	/* if this function is called w/o result, submit the call. otherwise, this is the callback from submit_call_to_server */
	//alert('general_element_change ' + element + ' record ' + recordid);

	if (typeof result == 'undefined') {

		var val = '';

		/* now check if this is a multiselect element */
		if ($('#JS_' + element + '-0-multiselect').length) {

			/* walk through the options add create an array of it */
			var vals = [];

			/* the div in which multiselects reside, walk through all checked checkboxes */
			var i = 0;
			$('#JS_' + element + '-multiselect').find(':input:checked').each(function() {

				/* add value to array-to-submit */
				vals[i] = $(this).val();
				i++;
			});

			/* create a JSONified version of the array */
			val = JSON.stringify(vals);

		} else {
			/* it's a normal field, or at least we need the plain value */
			val = $('#JS_' + element).val();
		}

		/* assemble data to post to server */
		var data = 'method=general_element_change&fqn=' + element + '&recordid=' + recordid + '&value=' + encodeURIComponent(val);

		/* add all field values, reloads need them */
		$('[id^="JS_"]').each(function() {

			if ($(this).attr('name') != "recordid" && $(this).attr('name') != "fqn" && $(this).attr('name') != "value" && typeof $(this).attr('name') !== 'undefined' )
			{
				/* add value to array-to-submit */
				data += '&' + $(this).attr('name') + '=' + encodeURIComponent($(this).val());
			}
		});

		/* submit call to server */
		/* alert(data); */
		/* DEBUG */
		submit_call_to_server(data, 'general_element_change', 'async');

	} else {

		if (result["result"] == "ok") {

			/* remove any error markers */
			/* something is wrong. place error marker. */
			if ($('#JS_' + element + '-multiselect').length) {
				remove_error(element + '-multiselect');
			} else {
				remove_error(element);
			}

			/* update the new value in the form */
			if (result["corrected_value"] != "") {
				if (result["corrected_value"] == "{{{emtpy}}}") {
					result["corrected_value"] = "";
				}
				$('#JS_' + element).val(result["corrected_value"]);
			}

		} else {

			/* little red flash. development only i think */
			// flash(element);

			/* something is wrong. place error marker. */
			if ($('#JS_' + element + '-multiselect').length) {
				set_error(element + '-multiselect');
			} else {
				set_error(element);
			}

		}
		/* and display a message */
		if (result["message"] != null && result["message"].length > 0) {

			/* the server js-encodes the message before sending it */
			alert(result["message"]);
		}

		/* re-enable the element */

		/*
		$('#JS_' + element ).prop('readOnly', false);
		*/

		/* update all elements pointing to the field we just updated
		var element			the element name of the updated field
		var recordid		the concerning record-id
		var target_table	array of tables to which a refer field was updated ("t12")
		var new_value		the new record in target_table where element[recordid] now refers to

		Important: the target_table variable is an array!

		/* compose array containing only myself and add it to the update elements */
		x = element.split('-');
		realfqn = x[0] + '-' + x[1];
		if (result["update_elements"].prop && result["update_elements"].prop.constructor !== Array) {
			result["update_elements"] = [];
		}
		result["update_elements"].push(realfqn);

		var elements_to_reload = [];

		/* if the updated field was a reference, IFields following the same path can be updated so they represent the same value */
		if (result["update_elements"] != null) {

			/* change all fields referring over a path which was just changed */

			update_all_elements_pointing_to_table_of_changed_reference_field(element, recordid, result["update_elements"], result["corrected_value"]);

			for (index = 0; index < result["update_elements"].length; ++index) {

				/* if the value contains an fqn it's a common field to update. if it's only a table it's a reference update */
				x = result["update_elements"][index].split('-');
				table = x[0];
				field = x[1];

				 /* alert('update table ' + table + ' field ' + field);  */

				 if (table != '' && field != '' && (result["update_elements"][index] != realfqn)) {

					/* build up an array with elements to reload */
					elements_to_reload.push(result["update_elements"][index]);

				}
			}
		}

		/* reload all elements at once */
		reload_element(elements_to_reload, recordid);

		/* update all calculations if the field was autosaved */
		if (result["autosaved"] == 1) {
			/* reload_calculations(element, recordid); */
			 reload_lists_and_calculations_in_all_parents();

		}

		/* update all clientside fields having client side validations */

		if (typeof hideshowallelements === 'function') {
			hideshowallelements();
		}
	}
}
/*
function update_comment_field



*/
function update_comment_field(fqn, recordid, result) {

	/* if this function is called w/o result, submit the call. otherwise, this is the callback from submit_call_to_server */

	if (typeof result == 'undefined') {

		/* assemble data to post to server */
		var data = 'method=update_comment_field&fqn=' + fqn + '&recordid=' + recordid;

		/* submit call to server */
		submit_call_to_server(data, 'update_comment_field', 'async');

	} else {

		$('#comment_' + fqn).html(result["result"]);

	}
}
/*
function update_all_elements_pointing_to_table_of_changed_reference_field

Walks through all ifields on the page to check if they refer to a table to which
the reference was just updated.

A form can contain ifields which refer to a record in a higher table.
When the user changes this reference record, all ifields using the
same higher table should be updated (refreshed).

var element			the element name of the updated field
var recordid		the concerning record-id
var target_table	*array* of tables to which a refer field was updated ("t12")
var new_value		the new record in target_table where element[recordid] now refers to
*/
function update_all_elements_pointing_to_table_of_changed_reference_field(element, recordid, target_table, new_value) {



	/*
	 *alert('Update all refs. Element: ' + element + ' record ' + recordid + ' target_table ' + target_table + ' new value ' + new_value);
		let op: hier ook nog door alle "normale" waarden lopen! dus bv. @t1-f12-t2-f134@ moet ook gereplaced worden.
	*/
	var data = 'method=reload_ifield';
	var some_data = false;

	$('.ifield-enclosing-div').each(function() {

		/* aantal element max 4 en alleen als callingelement = fqpath[1] */
		var me = this.id.replace('ifield-enclosing-div-', '');
		var parkthis = this;

		var el = me.split('-');

		/* alert('Update all refs to ' + target_table + ' current found ' + this.id + ' new value ' + new_value); */
		var tables_check = target_table; //.split(',')

		$.each(target_table, function(key, value) {

			var target_table = value;
			// alert('compare ' + el[0] + ' to ' + target_table);

			if (el[0] == target_table || el[0] + '-' + el[1] == target_table) {

				//alert('Add ' + me + ' to requested');
				data += '&fqn[' + me + ']=' + new_value;
				some_data = true;

			} else {
				/* alert('ref ' + el[2] + ' not on target'); */
			}
		});

	});

	if (some_data) {
		submit_call_to_server(data, '', 'async');
	}
}
function update_all_elements_pointing_to_table_of_changed_reference_field_callback(element, recordid, result) {

	/*
	 *console.log('got back element ' + element + ' for recordid ' + recordid);
	 */

	/* warning: field "element" now also contains the recordid */



	if (result["message"] != "" && typeof result["message"] != 'undefined') {
		alert(result["message"]);
	}

	$.each(result['result'], function (key, value) {

		//$('#ifield-enclosing-div-' + key).css('background-color', 'red');
		$('#ifield-enclosing-div-' + key).html(value);
	});

}

/*
var data				the data to post
var callback			the function to return the result to (must be known in function)
*/
function submit_call_to_server(data, callback, sync_or_async) {

	var async = false;

	if (sync_or_async == 'async')
	{
		async = true;
	}

	/* add the id of the cached current state, if it exists */
	if ($('#JS_current_state_id').val() > 0)
	{
		data += '&current_state_id=' + $('#JS_current_state_id').val();
	}

	/*
	 * alert('Load  ' + data);
	 *
	 */

	if (!async) {

		return $.parseJSON($.ajax({
			type: "POST",
			url: "record-ajax.php",
			async: false,
			data: data
		}).responseText);

	} else {

		$.ajax({
			type: "POST",
			url: "record-ajax.php",
			async: async,
			data: data,
			success: function(data) {
				/* JSON data array layout:

				[0] method
				[1] fqn element
				[2] recordid
				[3] array of return values:

					["result"]				ok/nok/readonly
					["message"]				message to display
					["update_elements"]		form elements to update
					["autosaved"]			1|0 autosaved or not
					["corrected_value"]		the value after validation

				*/

				arr = $.parseJSON(data);
				if (callback == 'general_ifield_change') {
					general_ifield_change(arr[1], arr[2], '', '', '', '', arr);
				} else if (arr[0] == 'general_element_change') {
					general_element_change(arr[1], arr[2], arr[3]);
				} else if (arr[0] == 'update_comment_field') {
					update_comment_field(arr[1], arr[2], arr[3]);
				} else if (arr[0] == 'reload_element') {
					reload_element(arr[1], arr[2], arr[3]);
				} else if (arr[0] == 'reload_ifield') {
					update_all_elements_pointing_to_table_of_changed_reference_field_callback(arr[1], arr[2], arr[3]);
				} else if (arr[0] == 'reload_calculations') {
					reload_calculations_callback(arr[3]["result"], arr[3]["message"]);
				} else {
					alert('Unknown callback function ' + arr[0] + '. Quitting.');
				}
			},
			error: function() {
				ShowConnectionError();
			}
		});
	}
}
/*
function put_id_in_child

Puts a newly added record id in the refer field element of the child after adding.

var elemement_in_child		the element name of the field in the child window to update
var value					the value (raw) to put in there
var textvalue				the human readable value to put in JS_ + element name + _ts

*/
function put_id_in_child(element_in_child, value, textvalue) {

	/* get object of target field to update */
	selectfield = WhosMyDaddy().document.getElementById('JS_' + element_in_child);

	/* update that field to the newly created record id as given */
	selectfield.value = value;

	/* check to see if there is a _ts field. _ts means "to show" and it contains a more readable value. the server passed this along in textvalue. */
	if (WhosMyDaddy().document.getElementById('JS_' + element_in_child + 'ts')) {
		/* if it exists, update it */
		WhosMyDaddy().document.getElementById('JS_' + element_in_child + 'ts').innerHTML = escapeHTML(textvalue);
	}

	/* check to see if there is a delete icon which was hidden before, but must be made visible now. */
	// if a field is required; this deleteicon will not exist
	// if a field is NOT required; it will exists but might be hidden (if no value es currently selected)
	if (WhosMyDaddy().document.getElementById('blindrefdeleteicon' + selectfield.id)) {
		/* if it exists, show it */
		WhosMyDaddy().document.getElementById('blindrefdeleteicon' + selectfield.id).style.display = 'inline';
	}

	/* run the onchange event of the field we just updated */
	if (typeof WhosMyDaddy().document.getElementById('JS_' + element_in_child).onchange == 'function')
	{
		WhosMyDaddy().document.getElementById('JS_' + element_in_child).onchange();
	}
	/* close this window to return to child */
	CloseMe();
}


function execute_button(fqn, recordid) {


	/* assemble data to post to server */
	var data = 'method=validate_buttonclick&fqn=' + fqn + '&recordid=' + recordid;

	/* add all field values, reloads need them */
	$('[id^="JS_"]').each(function() {

		/* exclude buttons */
		if ($(this).attr('type') != 'button') {
			/* add value to array-to-submit */
			data += '&' + $(this).attr('name') + '=' + encodeURIComponent($(this).val());
		}
	});

	/* alert(data); */

	/* submit call to server. dont use [method, element, recordid, result] = ...etc because <IE9 doesn't support that */
	x = submit_call_to_server(data, '', 'sync');
	method		= x[0];
	element		= x[1];
	recordid	= x[2];
	result		= x[3];

	if (result['message'] != '') {

		alert(result['message']);
	}

	if (result['result'] == 'ok') {

		/* alert('button click approved'); */
		/* buttons are validated before this event is called. it is safe to assume we can submit */
		formid = $('#JS_' + fqn).get(0).form.id;

		/* determine the id. a button is always an extra field, so this is safe. */
		xa = fqn.split('-');
		buttonid = xa[1].replace('f', '');

		/* place the button id in the hidden field */
		$("#JS_e_button").val(buttonid);

		$('#' + formid).submit();
	} else {
		/*
		 alert('button click NOT approved');
		 */

		HWID('MainContent');


	}
}

function set_error(formfieldid)	{
	/*
		adds the errorinfield class

		this function adds the class to all sorts of elements in can find, e.g.
		when $('#JS_' + formfieldid + '-checkboxdiv') works, it's fine, it won't
		be there if the field is a textarea. Not the cleanest way, but works fine
		because of the tolerance of jquery.

	*/

	/*
	debug
	alert('set error ' + formfieldid + $('#JS_' + formfieldid ).css("borderBottomStyle"));
*/

	/* the normal errorinfield class uses borders to hightlight. if an element has no border visible, we fall back to the errorinfieldnoborders class */
	var classtoset = 'errorinfield';
	var done = false;

/*
	if ($('#JS_' + formfieldid ).css("borderBottomStyle") != "solid") {
		if ($('#JS_' + formfieldid ).parent().css("borderBottomStyle") != "solid") {
			classtoset = 'errorinfieldnoborders';
		} else {
			alert('zet op parent');
			$('#JS_' + formfieldid ).parent().addClass(classtoset);
			done = true;
		}
	}
	*/
	if (!done) {
		/* a normal form element */
		$('#JS_' + formfieldid ).addClass(classtoset);

		/* a checkbox enclosing div */
		$('#JS_' + formfieldid + '-checkboxdiv').addClass(classtoset);

		/* a reference-to-table field */
		$('#JS_' + formfieldid + 'flexreficons').addClass(classtoset);

		/* a radio select enclosing div */
		$('#JS_' + formfieldid + '-radiodiv').addClass(classtoset);

		/* a multiselect enclosing div */
		$('#JS_' + formfieldid + '-multiselect').addClass(classtoset);
	}

}
function flash(formfieldid) {
	/*
		flashes an element
	*/
	$('#JS_' + formfieldid ).addClass('flash');
	setTimeout( function(){
		$('#JS_' + formfieldid ).removeClass("flash");
	}, 1000)	// Timeout must be the same length as the CSS3 transition or longer (or we'll mess up the transition)
}
function remove_error(formfieldid) {
	/*
		removes the errorinfield class
	*/
	$('#JS_' + formfieldid ).removeClass('errorinfield');
	$('#JS_' + formfieldid + 'flexreficons').removeClass('errorinfield');
	$('#JS_' + formfieldid + '-radiodiv').removeClass('errorinfield');
	$('#JS_' + formfieldid + '-multiselect').removeClass('errorinfield');
}
function find_field_value(fqn, record) {

	if ( $('#JS_' + fqn + '-' + record).length) {
		val = $('#JS_' + fqn + '-' + record).val();
	} else if ( $('#--' + fqn + '--' + record + '-UPV').length) {
		val = $('#--' + fqn + '--' + record + '-UPV').html();
	} else {
		val = get_field_value(fqn, record);
	}
	return val;
}
function get_field_value(fqn, recordid) {
	/*
	 * gets a single, raw value from the server. at the time of writing only in use for hideandshowjavascript.
	 * hideandshowjavascript will try to find the value on the form itself, but if it's not there, it will fetch it
	 * from the server in a desperate attempt to validat the conditions.
	 * /

	/* assemble data to post to server */
	var data = 'method=get_field_value&fqn=' + fqn + '&recordid=' + recordid;

	x = submit_call_to_server(data, '', 'sync');
	method		= x[0];
	element		= x[1];
	recordid	= x[2];
	result		= x[3];
	//alert('Got ' + result["result"] + ' from server');
	return(result["result"]);
}
function get_form_of_element(element_id) {
	/*
		this function fetches the form object of an element by going up the DOM
	*/
	node = document.getElementById(element_id);

	while (node.nodeName != "FORM" && node.parentNode) {
		node = node.parentNode;

	}

	return(node);
}

var last_list_url = '';
var last_list_title = '';
var last_load_is_list = false;
function refresh_contents_of_div(element_id, url, data, show_wait_image_div, method, liststoreload, is_list, callback_success) {

	/* check if our element exists */
	if ($('#' + element_id).length) {

		/* show the spinner if required */
		if (show_wait_image_div) {
			SWID(element_id);
		}

		/* default method is POST unless statet otherwise */
		if (method != "GET") {
			method = "POST";
		}

		/* submit url to server (async) */
		$.ajax({
			type: method,
			url: url,
			data: data,
			// otherwise IE9 will cache ajax responses (only if type == GET)
			cache: false,
			success: function(data) {
				HideConnectionError();

				/* set innerhtml of element with returned content */
				$('#' + element_id).html(data);

				/* run jqueryinit */
				jQueryInit();

				/* initialize datepicker class in case some element carrying this class was returned */
				InitDatePicker();

				/* see if there are lists to reload */
				if (typeof liststoreload != 'undefined' && liststoreload != '') {

					/* list names can be separated with commas */
					allliststoreload = liststoreload.split(',');

					/* walk through lists and refresh them */
					for (var i=0; i < allliststoreload.length; i++) {
						ListReload(allliststoreload[i], true, true);
					}
				}

				if (is_list)
				{
					last_list_url = url;
					last_list_title = $("h1:first").text().trim();
				}
				last_load_is_list = is_list;

				if (callback_success != '') {
					eval(callback_success);
				}
			},
			error: function() {
				if (show_wait_image_div) {
					HWID(element_id);
				}
				ShowConnectionError();
			}
		});

	} else {
		/* show a message */
		alert('element ' + element_id + ' does not exist');
	}
}
function reload_lists_and_calculations_in_all_parents(first_listid) {
	/*
		this function cycles through all parents. if it finds a form (JS_concerningtable and JS_concerningrecord) it will
		reload all calculations on that form. it also looks for #JS_ParentAjaxHandler element and if
		present, it will reload the parent's parent window with the found list id
	*/

	/* startpoint */
	var win = window;
//	alert('reload_lists_and_calculations_in_all_parents list id : ' + first_listid);

	/* if found, reload the list in the parent of this parent window */
	if (typeof first_listid != 'undefined')
	{
		/* normal action in normal popup. this reloads lists */
		if (WhosMyDaddy().$('#' + first_listid + 'divform').length)
		{
			//alert('call 1');
			WhosMyDaddy().ReloadAjaxElement(first_listid, '', true);
		}

		/* check if there is a div with this name - e.g. a calendar */
		if (WhosMyDaddy().$('#' + first_listid).hasClass('cal'))
		{
			//alert('call 2');
			WhosMyDaddy().reloadcal(first_listid, WhosMyDaddy().$('#' + first_listid + 'view').val(), WhosMyDaddy().$('#' + first_listid + 'startdate').val(), '', WhosMyDaddy().$('#' + first_listid + 'cals_to_show').val());
		}

	} else {

		/* in this case, we just bluntly reload all lists on the parent. expensive, but ok */
		WhosMyDaddy().$("input[name='listid']").each(function() {
			//alert('call 4');
			WhosMyDaddy().ReloadAjaxElement(this.value, '', true);
		});

	}

	/* try to find table and recordid */

	var fqn = $('#JS_concerningtable').val() + '-timestamp_last_change';
	var recordid = $('#JS_concerningrecord').val();

	/* if found, reload all calculations */
	if (fqn != '-timestamp_last_change' && recordid > 0)
	{
		//alert('call 5');

		/* skip lists */
		reload_calculations(fqn, recordid, true);
		/*
			alert('reload ' + fqn + ' for record ' + recordid);
		*/

	}

	while (win.WhosMyDaddy() && win.parent != win) {
		try {

			/* go to the parent */
			win = win.WhosMyDaddy();

			/* see if a parentajaxhandler elements exists */
			var ah = win.$('#JS_ParentAjaxHandler').val();

			/* if found, reload the list in the parent of this window */
			if (ah != '')
			{
				// alert('reload ' + ah);
				/* this checks the existence of a form element, it must contain 1 to reload the list. lists without
				   showing calculations should be reloaded for that's useless; they cannot be changed */

				if (win.WhosMyDaddy().$('#' + ah + 'reload_list_on_child_update').val() == 1)
				{
					//alert('call 6');
					win.WhosMyDaddy().ReloadAjaxElement(ah, '', true);
				}


			}


			/* try to find table and recordid */
			var fqn = win.$('#JS_concerningtable').val() + '-timestamp_last_change';
			var recordid = win.$('#JS_concerningrecord').val();

			/* if found, reload all calculations */
			if (fqn != '-timestamp_last_change' && recordid > 0)
			{
				//alert('call 7');
				/* skip lists */
				win.reload_calculations(fqn, recordid, true);

			}



			}catch(e) {
			}
	}


}

function request_notification_permission() {
	var Notification = window.Notification || window.mozNotification || window.webkitNotification;

	if (Notification.permission == 'denied') {
		Notification.requestPermission(function (permission) {
		// //console.log(permission);
		});
	}
}

function show_notification(title, msg) {

	request_notification_permission();
	var instance = new Notification(
		title, {
			body: msg
		}
	);

	instance.onclick = function () {
		// Something to do
	};
	instance.onerror = function () {
		// Something to do
	};
	instance.onshow = function () {
		// Something to do
	};
	instance.onclose = function () {
		// Something to do
	};


	return false;

}

/* show_notification('Interleave', 'Welkom in Interleave. Heel veel plezier he!');*/

function reload_calendar(args) {
	alert(args);
}

/*
function fix_url
arguments: url


checks whether an url starts with an http:// or https:// protocol name. if not, http:// is added

*/

function fix_url(url) {

	var https = url.substring(0, 8).toLowerCase();
	var http  = url.substring(0, 7).toLowerCase();

	if (http == 'http://' || https == 'https://')
	{
		return(url);
	} else {
		return('http://' + url);
	}
}

function microtime(get_as_float) {
	var unixtime_ms = (new Date).getTime();
	var sec = Math.floor(unixtime_ms/1000);
	return get_as_float ? (unixtime_ms/1000) : (unixtime_ms - (sec * 1000))/1000 + ' ' + sec;
}

function invert_all_boxes_with_class(classname) {

	/* inverts all checkboxes of a specific class */

	$('.' + classname).each(function() { //loop through each checkbox
		if (this.checked) {
			this.checked = false;
		} else {
			this.checked = true;
		}

	});


}
function nl2br(str, is_xhtml) {
  //  discuss at: http://phpjs.org/functions/nl2br/
  // original by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
  // improved by: Philip Peterson
  // improved by: Onno Marsman
  // improved by: Atli ��r
  // improved by: Brett Zamir (http://brett-zamir.me)
  // improved by: Maximusya
  // bugfixed by: Onno Marsman
  // bugfixed by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
  //    input by: Brett Zamir (http://brett-zamir.me)
  //   example 1: nl2br('Kevin\nvan\nZonneveld');
  //   returns 1: 'Kevin<br />\nvan<br />\nZonneveld'
  //   example 2: nl2br("\nOne\nTwo\n\nThree\n", false);
  //   returns 2: '<br>\nOne<br>\nTwo<br>\n<br>\nThree<br>\n'
  //   example 3: nl2br("\nOne\nTwo\n\nThree\n", true);
  //   returns 3: '<br />\nOne<br />\nTwo<br />\n<br />\nThree<br />\n'

  var breakTag = (is_xhtml || typeof is_xhtml === 'undefined') ? '<br ' + '/>' : '<br>'; // Adjust comment to avoid issue on phpjs.org display

  return (str + '')
    .replace(/([^>\r\n]?)(\r\n|\n\r|\r|\n)/g, '$1' + breakTag + '$2');
}

/* CKEDITOR FILE SELECT FUNCTIONS */

// Helper function to get parameters from the query string.
function getUrlParam( paramName ) {
	var reParam = new RegExp( '(?:[\?&]|&)' + paramName + '=([^&]+)', 'i' );
	var match = window.location.search.match( reParam );

	return ( match && match.length > 1 ) ? match[1] : null;
}
// Simulate user action of selecting a file to be returned to CKEditor.
function returnFileUrl(fileid, pub, filename, repos) {

	var funcNum = getUrlParam( 'CKEditorFuncNum' );

	var fileUrl =  document.getElementsByTagName('base')[0].href + 'csv.php?fileid=' + fileid;
	if (pub) {
		fileUrl =  document.getElementsByTagName('base')[0].href + 'publicfile.php?filename=' + filename + '&repository=' + repos;
	}
	window.opener.CKEDITOR.tools.callFunction( funcNum, fileUrl, function() {
		// Get the reference to a dialog window.
		var dialog = this.getDialog();
		// Check if this is the Image Properties dialog window.
		if ( dialog.getName() == 'image' ) {
			// Get the reference to a text field that stores the "alt" attribute.
			var element = dialog.getContentElement( 'info', 'txtAlt' );
			// Assign the new value.
			if ( element )
				element.setValue( 'alt text' );
		}
		// Return "false" to stop further execution. In such case CKEditor will ignore the second argument ("fileUrl")
		// and the "onSelect" function assigned to the button that called the file manager (if defined).
		// return false;
	} );
	window.close();
}
function load_attribute_field(identifier, attribute, entity) {
	//refresh_contents_of_div(element_id, url, data, show_wait_image_div, method, liststoreload, is_list, callback_success) {
	refresh_contents_of_div('JS_attr_' + identifier + '-' + attribute + '-' + entity, 'attrib-assist.php', 'identifier=' + identifier + '&attribute=' + attribute + '&entity=' + entity + '&edit=1', false, 'GET', '', false, false);
}

/* END CKEDITOR FILE SELECT FUNCTIONS */
