<?php
/* ********************************************************************
* Interleave
* Copyright (c) 2001-2016 Atomos Applications The Netherlands - www.atomos.nl - info@atomos.nl - www.interleave.nl - info@interleave.nl
* Licensed under the GNU GPL. For full terms see http://www.gnu.org/
*
* This script contains record functions
*
* Check http://www.interleave.nl/ for more information
*
* BUILD / RELEASE :: 7.0.20161227
*
**********************************************************************
*/
function process_record_save($table, $record, $data, $current_record = array(), $access = '')
{

    /*
        process_record_save: the standard record add/update handler

        $table				DS tablename
        $record				Record id or empty for add
        $data				Associative array of data to change (could be a copy of $_POST)
        $current_record		The current record. If given empty, it will be fetched
        $access				"ok" only. If not set, it will be figured out.

        returns:

        array(
            [0]	success bool
            [1] recordid
            [2] array with messages
        )

    */

    /* array to hold return messages */
    $messages = array();

    $fields_to_update = array();
    $sql_fieldpart = array();
    $posted_fields = array();
    $sql_datapart = array();
    $field_todos = array();

    $to_journal = '';
    $to_datajournal = array();

    $update = false;
    $add = false;

    if (ipi($record)) {
        $update = true;
    } else {
        $add = true;
    }

    if ($access != 'ok' || (count($current_record) == 0 && ipi($record))) {

        /* $access should have been given to this function and it should be "ok" but if it's not given we have to check ourselves */

        $access = 'nok';

        if (!ipi($record) && $table != '' && $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_add']) {
            $access = 'ok';
        } elseif (ipi($record) && is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) {
            $fields_to_fetch = array();
            $selectallfields = false;
            $bypass_record_access_by_sql = false;
            foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
                $fields_to_fetch[] = $table.'-'.$field;
            }

            // load the current_record variable
            $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

            if ($current_record['type_of_access'] == 'ok') {

                /* nothing special, access granted, all fine */
                $access = 'ok';

                /* insert a lock in the lock table */
                $sql = 'INSERT INTO CRMentitylocks(lockby, lockon, locktable, timestamp_last_change) VALUES ('.$GLOBALS['USERID'].",'".mres($record)."','".mres($table)."', NOW()) ON DUPLICATE KEY UPDATE timestamp_last_change=NOW()";

                mcq($sql);
            } elseif ($current_record['type_of_access'] == 'readonly') {

                /* read-only, various reasons */
                $access = 'readonly';
            }
        } else {
            $messages[] = 'Unclear what to do.';

            return array(false, $record, $messages);
        }
    }
    $saveonadd = false;
    if ($add) {
        if (GetAttribute('flextable', 'CreateEmptyRecordOnAdd', str_replace('t', '', $table)) == 'Yes') {
            $saveonadd = true;
        }
    }
    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
        $dsname = $table.'-'.$field;
        $dsname_in_post = '';
        $message = '';
        $value = '';

        if (!ipi($record)) {
            $record = '_new_';
        }
        $dsname_in_post = $table.'-'.$field.'-'.$record;
        /* print 'Check for ' . $dsname_in_post . '<br>'; */

        if (isset($data[$dsname_in_post]) && $field != 'recordid') {

            /* we only allow setting timestamp_last_change when adding a new records on opening the form */
            if ($field != 'timestamp_last_change' || $saveonadd) {

                /* check if this field is writeable by this user. log error if it isn't */
                if (($current_record['type_of_access'] == 'ok' && $current_record['*'.$dsname] == 1) || $add) {

                    /* some corrections */
                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype'] == 'date') {
                        if ($current_record[$dsname] == '0000-00-00 00:00:00') {
                            $current_record[$dsname] = '';
                        }
                        if ($current_record[$dsname] == '0000-00-00') {
                            $current_record[$dsname] = '';
                        }
                    } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '' && $current_record[$dsname] == 0) {
                        $current_record[$dsname] = '';
                    }
                    if (array_key_exists($dsname_in_post, $data) && (trim($data[$dsname_in_post]) != $current_record[$dsname] || $add || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect'])) {

                        // Validate field input here

                        $result = 'nok';

                        /* check for multiselect and diary fields (strange ducks in the byte) */

                        if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'diary') {

                                /* diary additions can be submitted also (next to ajax) */

                                /* get current diary content. it's in the database as a serialized array (ouch) */
                                $current_diary = array();
                            if ($current_record[$dsname] != '') {
                                $current_diary = unserialize($current_record[$dsname]);
                            }

                            if ($data[$dsname_in_post] != '') {
                                /* add array with new value */
                                    $current_diary[] = array(date('U'), $GLOBALS['USERID'], $data[$dsname_in_post]);
                            }

                                /* compose stuff to put into the database */
                                $value = serialize($current_diary);

                            if (is_array(unserialize($value))) {
                                $result = 'ok';
                            }
                        } else {
                            $value = $data[$dsname_in_post];

                            if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)' || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date') {

                                /* reformat date fields */

                                $value = FormattedDateToNLDate($value);
                            } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date/time') {
                                /*
                                    do nothing. inconsistent, but ValidateFieldInputDataStructureOnly expects date/time fields to be user-formatted
                                */
                            }

                            /* validate the field value */
                            list($result, $message, $value) = ValidateFieldInputDataStructureOnly($dsname, $value, true, $record);

                            if ($result == 'ok') {

                                /* we don't support corrections on multiselect fields, so we use the original array, since it validated */
                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect']) {

                                    /* re-built the array with base64 encoded values (legacy) */
                                    $value = array();
                                    foreach ($data[$dsname_in_post] as $val) {
                                        $value[] = base64_encode($val);
                                    }

                                    /* serialize it, this is the way it has to be put into the database (legacy) */
                                    $value = serialize($value);
                                } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)') {

                                    /* the validator returns dutch formatted dates. for iso, we need to convert it to yyyy-mm-dd */
                                    $value = NLDate2INTLDate($value);
                                }
                            } else {
                                //$messages[] =  'ValidateFieldInputDataStructureOnly returned ' . $result . '';
                            }
                        }

                        if ($result == 'ok') {
                            if ($add) {
                                // Add to-do for triggers
                                /*
                                if (ipi(str_replace("f", "", $field))) {
                                    $field_todos[] = array("EFID" . str_replace("f", "", $field), $value);
                                }
                                */

                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '' && $value == '0') {
                                    /* when adding a new record, a reference field containing "0" means not set at all. as this is database default, we can safely ignore it here to make sure it isn't logged and no rules are kicked  */
                                        $value = '';
                                }

                                /*
                                     only set something if it's given. as we're ADDING a record, the current value is always empty
                                    also process fields which have a non-empty default value, but were posted empty
                                */
                                if ($value != '' || ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['defaultvalue'] != '' && $value == '')) {

                                    /* add to to-do query parts */
                                    $sql_fieldpart[] = '`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database'].'`';
                                    $sql_datapart[] = mres($value);

                                    /* add to-do for rules. store in array, the record number is unknown */
                                    /* DIT STOND EERST NET BUITEN DEZE IF LUS. GEVAARLIJKE CHANGE */
                                    $field_todos[] = array($table.'-'.$field, $value);

                                    $to_datajournal[] = array($value, $field);
                                    $to_journal .= 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                                        $to_journal .= $current_record[$dsname].':'.GetUserName($current_record[$dsname]);
                                    } else {
                                        $to_journal .= $current_record[$dsname];
                                    }

                                    $to_journal .= '] to [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                                        $to_journal .= GetUserName($value);
                                    } else {
                                        $to_journal .= $value;
                                    }

                                    $to_journal .= "]\n";

                                    /* we need to pass this array to add_default_values_for_fields to make sure they're skipped when adding default values */
                                    $posted_fields[] = $field;
                                    //print '<h2>Add ' . $field . ' to posted_fields because the value is "' . htme($value) . '" and the defaultvalue is "' . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]['defaultvalue'] . '</h2>';
                                } else {
                                    //print '<h2>Do NOT add ' . $table . '-' . $field . ' to posted_fields because the value is "' . htme($value) . '" and the defaultvalue is "' . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]['defaultvalue'] . '</h2>';
                                }
                            } elseif ($update) {

                                /* TODO: DEZE FUNCTIE LOGGED EEN ONTERECHTE CHANGE OP EEN DIARY VELD IN ALLE GEVALLEN OOK ALS DEZE NIET GEUPDATE IS! */

                                /* correct for proper comparison with getrecord output */
                                $cmpval = $value;
                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)' || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date') {

                                        /* reformat date fields */
                                        $cmpval = NLDate2INTLDate($value);
                                }

                                /* check if it was changed */
                                if ($cmpval != $current_record[$dsname]) {

                                    /* add to list of fields to update */
                                    $fields_to_update[$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database']] = $value;

                                    /* log a todo */
                                    AddToDo($table.'-'.$field, $value, $record, $current_record[$dsname]);

                                    /* log a message in the datajournal */
                                    DataJournal($record, $current_record[$dsname], $value, $field, ds_tablename_to_datajournal_tabletype($table));

                                    /* add text for journal */
                                    $to_journal .= 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from [';

                                    /* some formatting for the journal. this is quite expensive, but much more friendly for the user */
                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '') {
                                        $to_journal .= $current_record[$dsname].':'.get_refer_layout_for_record($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'], $current_record[$dsname]);
                                    } else {
                                        $to_journal .= $current_record[$dsname];
                                    }

                                    $to_journal .= '] to [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '') {
                                        $to_journal .= $value.':'.get_refer_layout_for_record($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'], $value);
                                    } else {
                                        $to_journal .= $value;
                                    }

                                    $to_journal .= "]\n";
                                }
                            }
                        } else {
                            $messages[] = 'Error in '.$field.'. Record not added or updated: '.$message.'.';

                            return array(false, $record, $messages);
                        }
                    }
                } else {
                    /* this is not fatal; the record was saved, but this field was excluded */
                    log_msg('ERROR: Field ['.$field."] was posted, but the user doesn't have access to write to this field.");

                    if (GetSetting('DevelopmentOrProduction') == 'Development') {
                        $messages[] = 'ERROR: Field ['.$field."] was posted, but the user doesn't have access to write to this field.";
                    }
                }
            }
        }
    }

    //qlog(INFO, string_r($field_todos));

    if ($update && $access == 'ok') {
        $sql = 'UPDATE `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` SET ';
        $sqlparts = array();

        foreach ($fields_to_update as $field => $value) {
            $sqlparts[] = '`CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$field."` = '".mres($value)."' ";
        }

        $sql .= implode(',', $sqlparts);

        $sql .= ' WHERE '.$GLOBALS['UC']['PersonalDataStructure'][$table]['primary'].' = '.$record;

        if (count($sqlparts) > 0) {

            /* execute the query */
            mcq($sql);

            /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
            check_admin_changes_after_change($table);

            if ($to_journal != '') {
                journal($record, $to_journal, ds_tablename_to_journal_tabletype($table));
            }

            /* create to-do's for rules */
            AddToDo($table.'-changed', '', $record, '');

            $fields_to_fetch = array();
            $selectallfields = false;
            $bypass_record_access_by_sql = false;
            foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
                $fields_to_fetch[] = $table.'-'.$field;
            }

            /* reload the current_record variable */
            $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

            //$messages[] = $sql;
        } elseif ($update) {
            $update = false;

            /* in 99,9% of the cases this will be a form submit with not a single value *change*. not a real error */
        }
    } elseif ($add && $access == 'ok') {

        /* add timestamp_last_change field in case *nothing* was given to save (to prevent a crashing query) */
        if (!in_array('`timestamp_last_change`', $sql_fieldpart)) {
            $sql_fieldpart[] = '`timestamp_last_change`';
            $sql_datapart[] = date('Y-m-d H:i:s');
        }

        $sql = 'INSERT INTO `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` ('.implode(',', $sql_fieldpart).") VALUES('".implode("','", $sql_datapart)."')";

        mcq($sql);

//		print '<h1>' . $sql . '</h1>';

        $record = db_GetInsertID();

        /* journal this event */
        journal($record, 'Record created', ds_tablename_to_journal_tabletype($table));

        DataJournal($record, '-', '-', 'record created', ds_tablename_to_datajournal_tabletype($table));

        if ($to_journal != '') {
            journal($record, $to_journal, ds_tablename_to_journal_tabletype($table));
        }
        foreach ($to_datajournal as $dj) {
            DataJournal($record, '{{add}}',  $dj[0], $dj[1], ds_tablename_to_datajournal_tabletype($table));
        }

        /* add all default values */
        add_default_values_for_fields($table, $record, $posted_fields);

        /* we now know the record number, add todo's SHOULD NOT USE TODO'S - PROCESSRULES SHOULD BE CALLED DIRECTLY */
        foreach ($field_todos as $ftd) {
            AddToDo($ftd[0], $ftd[1], $record, '');
        }

        /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
        check_admin_changes_after_change($table);

        /* log todo for later processing of rules */
        AddToDo($table.'-added', '', $record, '');
    } else {

        /* no access; return false */
        return array(false, $record, $messages);
    }

    /* we changed a record; we have to unset the materialized record access subquery (cache) */
    unset($GLOBALS['PageCache']['php_subquery_materialization']);

    return array(true, $record, $messages);
}

function get_parsed_template($table, $record, $passthru, $rec, $force_autoform, $access)
{
    /*
     *   This function should be able to parse any template type. It fetches the template content
     *   and determines the type. The rest is done by get_parsed_template_with_given_template.
     */
    $template_contents = '';
    $template_type = '';

    /* let's see if this table has a template set */
    $form = $GLOBALS['UC']['PersonalDataStructure'][$table]['form-template-id'];

    /* this overrules the above */
    if (ipi($passthru['template_to_use'])) {
        /* using the ftu var, a form to use can be given which overrules */
        $form = $passthru['template_to_use'];
    }

    if ($force_autoform) {
        $template_contents .= ParseAutoForms('@AUTO_WITH_CHILD_TABLES@', $table);
        $template_type = 'TEMPLATE_HTML_FORM';
    } else {
        $template_type = GetTemplateType($form);

        if ($template_type != 'TEMPLATE_HTML_FORM' && $template_type != 'TEMPLATE_HTML' && $template_type != 'TEMPLATE_XML' && $template_type != 'TEMPLATE_REPORT_PDF' && $template_type != 'TEMPLATE_PLAIN') {
            return 'Unsupported template type';
        }
        if ($template_type == 'TEMPLATE_HTML_FORM') {
            $template_contents .= ParseAutoForms(GetTemplate($form), $table);
        } else {
            $template_contents .= GetTemplate($form);
        }
    }

    return get_parsed_template_with_given_template($table, $record, $passthru, $rec, $force_autoform, $access, $template_contents, $template_type);
}

function get_parsed_template_with_given_template($table, $record, $passthru, $rec, $force_autoform, $access, $template_contents, $template_type)
{
    global $lang;

    $locked = false;
    $lockedby = '';
    $html = '';

    if (!ipi($record) && $table != '') {
        $record = '_new_';
    }

    /* define what to do per type. this DOES NOT INFLUENCE THE ORDER IN WHICH THEY HAPPEN */

    $todo = array();
    switch ($template_type) {

        case 'TEMPLATE_HTML':
        case 'TEMPLATE_PLAIN':
        case 'TEMPLATE_REPORT_PDF':
        case 'TEMPLATE_XML':
            $todo = array('generic', 'php', 'modules', 'listtags', 'language', 'usergroup', 'languageif');
            if ($table != '' && $record != '') {
                $todo[] = 'data';
            }
        break;

        case 'DEFAULTVALUE':
            $todo = array('generic', 'php', 'modules', 'language', 'usergroup', 'languageif');
            if ($table != '') {
                $todo[] = 'data';
            }
        break;

        case 'TEMPLATE_HTML_FORM':
            $todo = array('generic', 'php', 'modules', 'listtags', 'language', 'usergroup', 'languageif');
            if ($table != '') {
                $todo[] = 'form';
                $todo[] = 'data';
            }
            if (ipi($record)) {
                $todo[] = 'listtags';
            }
        break;

        case 'TEMPLATE_DASHBOARD':
            $todo = array('generic', 'php', 'modules', 'listtags', 'usergroup', 'language', 'dashboard', 'languageif');
        break;

        case 'TEMPLATE_MENU':
            $todo = array('generic', 'php', 'modules', 'listtags', 'usergroup', 'language', 'languageif');
        break;

        default:
            log_msg('ERROR: get_parsed_template_with_given_template called with no or unkown tempate type.');
        break;

    }
    //DA($todo, true);

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {
        if (ipi($rec['record_locked_by'])) {
            $locked = true;
            $lockedby = $rec['record_locked_by'];
        }

        $save_after_validate = 'false';
        if ($passthru['BYAJAX']) {
            $save_after_validate = 'true';
        }
        //print 'debug rf.php:502<br>';
        //$save_after_validate = "true";
        $html = '';
        if (!$locked) {
            $html = '<script type="text/javascript">';
            $html .= PrintExtraFieldForceJavascript();
            $html .= '</script>';
        }
        $html .= '<div id="lostlockmsg" class="warnmsg hide_content">'.htme($lang['lostlock']).'</div><div id="form-'.$table.'-'.$record.'"><form class="recordform" novalidate autocomplete="off" enctype="multipart/form-data" id="'.$table;

        $html .= '-'.$record.'" method="post" action="" onsubmit="return general_form_submit(this.id, \''.$record.'\', '.$save_after_validate.');" onkeyup="LockCounter=1;">'.$template_contents;
    } else {
        $html = $template_contents;
    }

    if (in_array('generic', $todo)) {

        /*
        *
        * Handles:
        *
        * Mobile / desktop switcher
        * Time, user, userid, group, groupid tags
        * Calendar tags
        * Info tags
        * Inline add form tags
        * Default value tags
        * Template inclusions
        * Default Report Tags
        * User and group fields
        * Absolutely addressedd ifields
        *
        */

        $html = ParseTemplateGeneric($html);
    }
    if (in_array('usergroup', $todo)) {
        $html = StripGroupOnlyTags($html);
    }
    if (in_array('languageif', $todo)) {
        $html = StripLanguageOnlyTags($html);
    }
    /* replace language tags */
    if (in_array('language', $todo)) {
        // Replace language tags like [[save]] and [[efid12]]
        $html = ParseTemplateLanguageTags($html);
    }

    if (in_array('form', $todo)) {
        if ($access != 'readonly') {
            $html = StripROOnlyTags($html);
        } elseif ($access == 'readonly') {
            $html = StripRWOnlyTags($html);
        }
        if (ipi($record)) {
            $html = StripNewOnlyTags($html);
        } else {
            $html = StripExistingOnlyTags($html);
        }
    }

    /* we also support parsing templates not bound to a table */
    if (in_array('data', $todo)) {

        /* now parse the record */

        $html = ParseTemplateRecordData($html, $table, $record, $passthru['duplicate']);
        if (in_array('form', $todo)) {

            /* replace hidden fields. syntax: #HEFID500[1]# */
            $locrecord = '';
            if (!ipi($record)) {
                $locrecord = '_new_';
            } else {
                $locrecord = $record;
            }
            $html = preg_replace("/\#HEFID(\d+)\[(.*?)\]\#/", '<input type="hidden" id="JS_'.$table.'-f\\1-'.$locrecord.'" name="'.$table.'-f\\1-'.$locrecord.'" value="\\2">', $html);

            $savebtn = '';
            if ($access == 'ok') {
                $savebtn = '<input type="submit" id="JS_submitbutton" name="btnsubmit" value="'.htme($lang['save']).'">';
            }

            $html = str_replace('#SAVEBUTTON#', escape_interleave_characters($savebtn), $html);

            $delbtn = '';

            /* passthru['no_deletes'] is important when opening a record using the link in a ref-to-flex field. if a user would have a delete button (if the parent is a new record, no relation exists) evrything goes haywire */

            if ($record != '_new_' && $access == 'ok' && $passthru['no_deletes'] == '' && $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_delete']) {
                if (CheckFunctionAccess('AllowedToUseDescendingDeletes') != 'ok' && CheckIfRecordHasChilds($table, $record) && !is_administrator()) {
                    $delbtn = "<input type='button' disabled value='".htme($lang['delete'])."' name='delete_record_including_childs' class='deletebutton'>";
                } else {
                    $delbtn = "<input type='button' value='".htme($lang['delete'])."' name='delete_record_including_childs' onclick=\"PopFancyBoxDimensioned('','record.php?record=".$record.'&amp;table='.$table.'&amp;delete=1&amp;nonavbar=1&amp;ParentAjaxHandler='.htme($passthru['ParentAjaxHandler']).'&amp;fromlist='.htme($passthru['fromlist'])."&amp;deleting_myself=1', 300, 500);\" class='deletebutton'>";
                }
            }

            $html = str_replace('#DELETEBUTTON#', escape_interleave_characters($delbtn).' ', $html);
            $html = str_replace('#DELETEINCLUDINGCHILDSBUTTON#', escape_interleave_characters($delbtn).' ', $html);

            /* some variables we need to pass along */

            $pass_along = array();
            $pass_along['form_submitted'] = '0';
            $pass_along['e_button'] = '';
            $pass_along['concerningrecord'] = $record;
            $pass_along['concerningtable'] = $table;
            $pass_along['hash'] = randomstring(32);

            if (ipi($passthru['duplicate'])) {
                $pass_along['duplicate_of'] = $passthru['duplicate'];
            }

            foreach ($pass_along as $name => $value) {
                $html .= '<input type="hidden" id="JS_'.$name.'" name="'.$name.'" value="'.htme($value).'">';
            }

            $html .= '</form></div>';
        }

        if (stristr($html, '@FAMILYTREE@')) {
            $html = str_replace('@FAMILYTREE@', escape_interleave_characters(familytree($table, $record, $passthru['starttable'], $passthru['startrecordid'], $passthru['fromlist'])), $html);
        }
    }

    // Evaluate modules when they are part of this template
    if (in_array('modules', $todo)) {
        if (stristr($html, '@MODULE')) { // some reference to a module exists
            $x = GetModules();
            foreach ($x as $module) {
                if (stristr($html, '@MODULE'.$module['mid'].'@')) {
                    if ($table != '') {
                        $html = str_replace('@MODULE'.$module['mid'].'@', escape_interleave_characters(RunModule($module['mid'], false, true, $record, str_replace('t', '', $table))), $html);
                    } else {
                        $html = str_replace('@MODULE'.$module['mid'].'@', escape_interleave_characters(RunModule($module['mid'], false, true, '', '')), $html);
                    }
                    qlog(INFO, 'Module '.$module['mid'].' was ran and included in this template');
                }
            }
        }
    }

    // Replace list tags e.g. #LIST|table=t2#
    if (in_array('listtags', $todo)) {

        /* TODO: we moeten hier nog iets verzinnen waardoor waardeloze list tags (sublist maar nog geen record) niet getoond worden */

        /* replace the tags */
        $html = ParseTemplateListTags($html);

        /* parse inline-add tags */
        $html = ParseInlineAddFormTags($html);
    }

    // Replace report tags
    $html = ParseTemplateCreateDefaultReportTags($html);

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {

        // Inline add forms
        $template = ParseInlineAddFormTags($template);

        // Arrow tag
        $html = parse_arrow_tag($html, $table, $record);

        // The lock icon
        if (!$locked) {
            $html = StripLockedOnlyTags($html);
            $html = str_replace('#LOCKICON#', '', $html);
        } else {
            $html = str_replace('#LOCKICON#', '<span class="fa fa-lock"></span>', $html);
            $html = str_replace('@LOCKBY@', escape_interleave_characters(GetUserName($lockedby)), $html);
            $html = str_replace('@LOCKEDONLY@', '', $html);
            $html = str_replace('@ENDLOCKEDONLY@', '', $html);
        }

        // Report icon (only for MINV) -- LEGACY
        if (strpos($html, '#REPORTICON#') !== false) {
            $flextableid = '';
            $flextablerecord = '';
            if (ipi(str_replace('t', '', $table))) {
                $flextableid = str_replace('t', '', $table);
                $flextablerecord = $record;
            }

            $reporticon = "<a href=\"javascript:poplittlewindow('entityreport.php?FlexTableRecord=".$flextablerecord.'&amp;FlexTable='.$flextableid."')\">";
            $reporticon .= '<span class="fa fa-file-word-o"></span>';
            $reporticon .= '</a>';

            $html = str_replace('#REPORTICON#', escape_interleave_characters($reporticon), $html);
        }

        // The file box and list
        if (strpos($html, '#FILEBOX#') !== false) {
            $new = FileUploadBox($table, $record);
            if (ipi($record)) {
                $new .= DisplayFileList($record, $table);
            }
            $html = str_replace('#FILEBOX#', escape_interleave_characters($new), $html);
        }
        if (strpos($html, '#FILELIST#') !== false) {
            $replace = '';
            if (ipi($record)) {
                $replace = escape_interleave_characters(DisplayFileList($record, $table));
            }
            $html = str_replace('#FILELIST#', $replace, $html);
        }

        // The journal icon + popuplink

        // The journal icon. Don't show if not allowed.
        $journalicon = '';
        if ((CheckFunctionAccess('DenyJournalAccess') != 'ok' || is_administrator()) && $record != '_new_') {
            $journalicon = '<div id="journal-icon"><a title="Journal" href="journal.php?table='.$table.'&amp;recordid='.htme($record).'&amp;nonavbar=1" onclick="PopFancyBoxLarge(\'\', this.href); return false;">';
            $journalicon .= '<span class="fa fa-book"></span>';
            $journalicon .= '</a></div>&nbsp;';
    //jeroen 2016-02-14 deze use mag niet eens weten dat er zoiets is als een journal
    /*	} else {
            $journalicon = "<!-- user is not allowed to see the journal -->";*/
        }

        $html = str_replace('#JOURNALICON#', escape_interleave_characters($journalicon), $html);

        /* the lock message */

        if ($template_type == 'TEMPLATE_HTML_FORM') {
            if (!$locked) {
                /* remove text only for when locked */
                $html = StripLockedOnlyTags($html);
                /* insert javascript to update the lock on this record  */
                $html .= ReturnLockJavascript($record, $table);
            } else {
                /* leave lock message in place */
                $html = str_replace('@LOCKBY@', GetUserName($lockedby), $html);
            }
        }
    } // end if todo form

    if (in_array('dashboard', $todo)) {
        if (strstr($html, '#CLOCK#')) {
            $html = str_replace('#CLOCK#', ShowClock(), $html);
        }
    }

    if (in_array('php', $todo)) {
        if (stristr($html, '<?')) { // PHP exists in template
            $html = EvaluateTemplatePHP($html, false, $table, $record);
        }
    }

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {

        /* show inline and raw edit icons */
        if (($GLOBALS['SesMem']['InlineEditMode_InlineLinks'] == 'On' && is_administrator()) && ($template_type == 'TEMPLATE_HTML_FORM')) {

            // Show raw-edit icon
            if (ipi($record)) {
                $html = '<div class="showinline" style="position: relative"><div class="showinline InlineEditModeDiv" style="position: absolute;z-index:10;left:18px;top:-5px;"><a title="Edit this record raw" href="rawedit.php?MoreRaw&amp;record='.htme($record).'&amp;table='.htme($table).'"><span class="fa fa-wrench"></span></a></div></div>'.$html;
            }
        }
    }

    return unescape_interleave_characters($html);
}

function process_open_todos($table, $record)
{
    /*
        This function gets all todo's for rules (using fully qualified onchange field tags) and processes them. It ignores todo's made for triggers.
    */
    //trace("process_open_todos($table, $record)");
    $sql = "SELECT * FROM `CRMtodo` WHERE `CRMtodo`.`user`='".mres($GLOBALS['USERID'])."' AND LEFT(`CRMtodo`.`onchange`, ".strlen($table).") = '".htme($table)."' AND `CRMtodo`.`worker_hash` = '' AND `CRMtodo`.`eid`='".$record."'";

    $Todo = db_GetArray($sql);

    $sql = "SELECT COUNT(*) FROM `CRMrules` WHERE `CRMrules`.`table`='".mres($table)."'";
    $rules = mcq($sql);

    if (count($Todo) > 0 && $rules > 0) {
        if (!is_array($GLOBALS['SystemDataStructure'])) {
            $GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
        }
        // SLORDIG en memory intensief; Aan Jeroen vragen of dit netter kan.

        // Bypass PDS security by overwriting it to SDS
        $PDSPARK = $GLOBALS['UC']['PersonalDataStructure'];
        $GLOBALS['UC']['PersonalDataStructure'] = $GLOBALS['SystemDataStructure'];

        /*
            To prevent ProcessRules from having to get_records over and over we get it here.
        */

        /* determine the fields to select by checking ruleactions, ruleconditions and ruleactionconditions */
        $sql = "SELECT `CRMruleconditions`.`field` FROM `CRMruleconditions` WHERE `CRMruleconditions`.`field` LIKE '".$table."-%' UNION SELECT `CRMruleactionconditions`.`field` FROM `CRMruleactionconditions` WHERE `CRMruleactionconditions`.`field` LIKE '".$table."-%' UNION SELECT `CRMruleactions`.`field` FROM `CRMruleactions` WHERE `CRMruleactions`.`field` LIKE '".$table."-%'";

        /* now add all primaries to the fields to fetch */
        $fields_to_fetch_tmp = db_GetFlatArray($sql);
        $fields_to_fetch = array();

        foreach ($fields_to_fetch_tmp as $ftf) {

            /* add current always */
            $fields_to_fetch[] = $ftf;

            $elements = explode('-', $ftf);
            $element_table = $elements[count($elements) - 2];

            /* last table is not our own table, add primary */
            if ($element_table != $table) {
                unset($elements[count($elements) - 1]);
                $path = implode('-', $elements).'-'.$GLOBALS['UC']['PersonalDataStructure'][$element_table]['primary'];
                if (!in_array($path, $fields_to_fetch)) {
                    $fields_to_fetch[] = $path;
                }
            }
        }

        $selectallfields = false;
        $bypass_record_access_by_sql = false;

        $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

        // Put PDS back
        $GLOBALS['UC']['PersonalDataStructure'] = $PDSPARK;
        unset($PDSPARK);

        /* create a unique worker hash to put in database to mark records to process. (this can be made simpler; this seems expensive, nothing else relies on this format) */
        $worker_hash = md5(rand().rand().date('U').$GLOBALS['USERNAME'].$GLOBALS['USERID']).md5(rand());

        /* update the table to mark (lock) records for processing */
        $sql = "UPDATE `CRMtodo` SET `CRMtodo`.`worker_hash` = '".$worker_hash."' WHERE `CRMtodo`.`user`='".mres($GLOBALS['USERID'])."' AND LEFT(`CRMtodo`.`onchange`, ".strlen($table).") = '".htme($table)."' AND `CRMtodo`.`eid`='".$record."' AND `CRMtodo`.`worker_hash` = ''";
        mcq($sql);

        /* next, select the records we just marked. the order by clause is important; further down we fetch the records one by one */
        $sql = "SELECT * FROM `CRMtodo` WHERE `CRMtodo`.`worker_hash`='".$worker_hash."' ORDER BY `CRMtodo`.`eid`";
        $Todo = db_GetArray($sql);

        /* array to hold todo-id's which were processed and can be deleted */
        $Todelete = array();

        /* loop through all todo's */
        foreach ($Todo as $do) {
            if ($do['onchange'] == $table.'-added') {
                /* an added record */
                //trace("Processrules($table, $record, '', \"added\", " . $do['to_value'] . "");
                Processrules($table, $record, '', 'added', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } elseif ($do['onchange'] == $table.'-changed') {
                /* a changed record */
                Processrules($table, $record, '', 'changed', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } elseif ($do['onchange'] == $table.'-deleted') {
                /* a deleted record */
                //trace("Processrules($table, $record, '', \"deleted", " . $do['to_value'] . "");
                Processrules($table, $record, '', 'deleted', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } else {

                /* a generic change-of-a-field */
                /* get the field */
                list($ignore, $todofield) = explode('-', $do['onchange']);
                unset($ignore);

                /* check if the field is in the data structure */
                if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$todofield])) {

                    //trace("Processrules($table, $record, \"$todofield\", \"changed\", " . $do['to_value'] . ");");
                    Processrules($table, $record, $todofield, 'changed', $do['to_value'], $current_record);
                    $Todelete[] = $do['todoid'];
                }
            }
        }

        /* delete the processed todo's */
        if (count($Todelete) > 0) {
            $sql = 'DELETE FROM `CRMtodo` WHERE `CRMtodo`.`todoid` IN('.implode(',', $Todelete).')';
            mcq($sql);
        }
    } else {
        //		trace("niks te doen hoor");
    }

    // old code copy for processing triggers
    //jeroen: 2016-01-06 dit doet volgens mij niks meer. weg?
    $st = '';
    if (ipi(str_replace('t', '', $table))) {
        $st = str_replace('t', '', $table);

        $Todo = GetTodos($record);
        //DA($Todo);

        foreach ($Todo as $do) {
            if ($do['onchange'] == 'FlexTable'.$st.'-Change' || $do['onchange'] == 'FlexTable'.$st.'-Add' || $do['onchange'] == 'FlexTable'.$st.'-AnyUpdate' || $do['onchange'] == 'FlexTable'.$st.'-Delete' || GetExtraFieldTableType(str_replace('EFID', '', $do['onchange'])) == $st) {
                DropTodos($record, $do['onchange']);
            } else {
            }
        }
    }
}

function update_field_value($fqn, $recordid, $value, $bypass_security = false, $dont_kick_rules = false, $journal_addition = '(autosave)')
{
    /*
        This function updates a single field to a new value. The value must be given in RAW form; exactly in the way it is supposed to be put into the database!
    */
    $messages = '';
//	print "update_field_value($fqn, $recordid, " . string_r($value) . ", $bypass_security=false, $dont_kick_rules";

    if (!ipi($recordid)) {
        $messages .= "F: Empty record id\n";
        log_msg('ERROR: update_field_value called with empty record id');

        return $messages;
    } else {
        list($table, $field) = explode('-', $fqn);

        /* get the record, but only for the field we're looking for ($fqn) */
        $current_record = getrecord($table, $recordid, false, $bypass_security, array($fqn));

        if (($current_record['type_of_access'] == 'ok' && $current_record['*'.$fqn] == 1) || $bypass_security) {
            if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect']) {
                $submitted_value = $value;
                $value = array();
                foreach ($submitted_value as $val) {
                    $value[] = base64_encode($val);
                }
                $value = serialize($value);
            }

            /* if the database value of this field is not equal to the submitted value ... */

            if ($current_record[$fqn] != $value) {

                /* ... update it */
                $sql = 'UPDATE `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database']."` = '".mres($value)."' WHERE `CRM".$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['primary'].'` = '.$recordid;

                mcq($sql);

                /* journal this event in both journals */
                DataJournal($recordid, $current_record[$fqn], $value, $field, ds_tablename_to_datajournal_tabletype($table));

                $journalvalue_to = $value;
                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                    $journalvalue_to = $journalvalue_to.':'.GetUserName($journalvalue_to);
                }
                $journalvalue_from = $current_record[$fqn];
                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                    $journalvalue_from = $journalvalue_from.':'.GetUserName($journalvalue_from);
                }
                journal($recordid, 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from ['.$journalvalue_from.'] to ['.$journalvalue_to.'] '.$journal_addition, ds_tablename_to_journal_tabletype($table));

                if (!$dont_kick_rules) {

                    /* add to-do's for rules to maybe process rules later */
                    AddToDo($fqn, $value, $recordid, $current_record[$fqn]);
                    AddToDo($table.'-changed', '', $recordid, '');

                    /* add old-style trigger todo's */
                    if (ipi(str_replace('t', '', $table))) {
                        AddToDo('FlexTable'.str_replace('t', '', $table).'-AnyUpdate', '', $recordid);
                        AddToDo('FlexTable'.str_replace('t', '', $table).'-Change', '', $recordid);
                    } elseif ($table == 'users') {
                        AddToDo('user_change', '', $recordid); // Add entity change to todo list if it wasn't there
                        AddToDo('user_AnyUpdate', '', $recordid); // Add entity change to todo list if it wasn't there
                    }

                    /* add todo's for triggers on fields */
                    if (ipi(str_replace('f', '', $field))) {
                        AddToDo('EFID'.str_replace('f', '', $field), $value, $recordid);
                    }
                }

                /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
                check_admin_changes_after_change($table);
            //	$messages .= "D: " . string_r(db_GetArray("SELECT * FROM CRMtodo"));
            } else {
                // $messages .= "Field value not changed, not saving anything";
            }
        }

        return $messages;
    }
}
function is_associative_array($array)
{
    /* check if there are strings */
    $a = (bool) count(array_filter(array_keys($array), 'is_string'));
    if ($a) {
        /* if not, check if keys are sequential */
        return array_keys($array) !== range(0, count($array) - 1);
    }

    return false;
}
function get_selectfield_options($fqn, $recordid)
{
    global $user;
    global $pass;
    global $host;
    global $database;
    /* ignore undeclared-and-unused-var-checker.php about this $db global */
    global $db;

    /* array to hold additional classes to return */
    $classes = array();

    $is_assoc = false;
    /*
        This function returns an associative array with options for a select field. Source is datastructure, PHP code, or SQL query
    */
    list($table, $field) = explode('-', $fqn);

    /* the id of the field */
    $field_id = str_replace('f', '', $field);

    if (GetAttribute('extrafield', 'RepopulateByAjax', $field_id) == 'Yes') {
        $classes[] = 'calculation';
    }

    /* array to hold the eventual options */
    $options_list = array();

    /* fetch the complete field row - we need it because not everything is in the datastructure */
    $extrafield = db_GetRow("SELECT * FROM `CRMextrafields` WHERE `CRMextrafields`.`id`='".$field_id."'");

    /* start checking field variants. start with a field populated by PHP code */
    if (substr($extrafield['options'], 0, 18) == '%POPULATE_BY_CODE%') {

        /* an PHP code evaluation should return an array called $result being an array, could be assiociative and plain */
        $result = array();

        /* get the code */
        $code = str_replace('%POPULATE_BY_CODE%', '', $extrafield['options']);

        /* code almost always contains tags, parse it */
        $code = unescape_interleave_characters(ParseTemplateRecordData($code, $table, $recordid));

        /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
        $flextableid = '';
        $flextablerecord = '';
        $eid = '';

        /* set some vars for backwards compatibility */
        if (ipi(str_replace('t', '', $table))) {

            /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
            $flextableid = str_replace('t', '', $table);
            $flextablerecord = $recordid;
        }
        /* evaluate the code */
        @eval($code);

        /* loop through the results and check if it's 1 or 2 columns and put it into options_list */
        foreach ($result as $row) {
            //if ($row[1] != "") {
            if (array_values($row)[1] != '') {
                $is_assoc = true;
                $options_list[$row[0]] = $row[1];
            } else {
                $options_list[] = $row;
            }
        }
    } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'SQL Query') {

        /* a field which options are populated by the output of an SQL query. If it returns 1 column than value=text; two columns means [0]=value [1]=text */

        /* the database name for the query is stored in a weird way: a serialized array */
        $o = @unserialize($extrafield['options']);
        $dbase = $o[0];
        $query = $o[1];

        /* replace tags in the query */
        $query = ParseTemplateGeneric($query);
        $query = ParseTemplateRecordData($query, $table, $recordid);
        $query = unescape_interleave_characters($query);
        $query = str_replace("\n", ' ', $query);

        if ($dbase != '') {
            /* check to see if the database name also has a host (syntax: database_name@hostname) */
            if (strpos($dbase, '@') !== false) {

                /* it does, connect to it */
                $tmp = explode('@', $dbase);
                $dbase = $tmp[0];
                $tmphost = $tmp[1];
                $db = db_only_connect($user[$GLOBALS['repository']], $pass[$GLOBALS['repository']], $tmphost);
            }

            /* select the given database */
            if (!db_only_select($dbase)) {
                /* couldn't connect; put error INSIDE field options. not a nice way to do this but it's clear to the user */
                $options_list[0] = 'Could not select database ['.$dbase.']';
            }
        }

        $res = array();
        /* run the query */
        if (trim($query) != '') {
            $res = db_GetArray($query);
        }
        /* loop through the results and check if it's 1 or 2 columns and put it into options_list */
        foreach ($res as $row) {
            if ($row[1] != '') {
                $is_assoc = true;
                $options_list[$row[0]] = $row[0].': '.$row[1];
                //print 'add key ' . $row[0] . ' value ' . $row[1] . '<br>';
            } else {
                $options_list[] = $row[0];
            }
        }

        if ($dbase != '') {
            /* connect back to our interleave database */
            $db = db_base_connect($user[$GLOBALS['repository']], $pass[$GLOBALS['repository']], $host[$GLOBALS['repository']], $database[$GLOBALS['repository']]);

            /* load environment. is this really nescessary?? */
            SwitchToRepos($GLOBALS['repository']);
        }
    } else {

        /* it's a normal options field - options are in the data structure, always non-associative */
        $options_list = $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['options'];
    }

    /* if the field properties say so, sort the array */
    if ($extrafield['sort'] == 'y') {
        natcasesort($options_list);
    }

    $return = array();

    /* clean up the options_list */
    // dit werkt nog niet lekker. een array met 4 => bla 8 => bla2 wordt onterecht opgevat als niet-associative. Als er *ergens* een string in staat gaat het wel goed.
    if (!is_associative_array($options_list) && !$is_assoc) {
        foreach ($options_list as $key => $value) {
            unset($key);
            $return[$value] = $value;
        }
    } else {
        $return = $options_list;
    }
    //DA($return, false);
    /* return the result */
    return array($return, $classes);
}

function get_elements_to_update_on_field_change($fqn)
{
    /*
        This function checks which other fields on a form need to be refreshed when a field is changed. This could
        be due to the attributes RefreshByAjaxOnChangeOfField and RepopulateByAjax, but also field requirement
        conditions could have a different result because of a change.
    */

    list($table, $field) = explode('-', $fqn);

    /* the array to hold the result */
    $result = array();
    //$result[] = "CALLED WITH $fqn";

    /* the field id of the field in the fields table */
    $field_id = str_replace('f', '', $field);

    /* get all field-ids (.entity) and CSV-lists of field (.value) of all fields. this query could be more specific */
    $sql = "SELECT `CRMattributes`.`entity`, `CRMattributes`.`value` FROM `CRMattributes` WHERE `CRMattributes`.`attribute` ='RefreshByAjaxOnChangeOfField' AND `CRMattributes`.`identifier` ='extrafield' AND `CRMattributes`.`value` != ''";
    $lor = db_GetArray($sql);

    /* loop through all fields having something set */
    foreach ($lor as $row) {

        /* explode the value (it's stored like 1,2,4) */
        $lof = explode(',', $row['value']);

        /* if the field which was updated is in this list, this field needs to be refreshed */
        if (in_array($field_id, $lof) || $lof[0] == 'All') {

            /* add this field to the result set */
            $result[] = get_ds_tablename_for_field($row['entity']).'-f'.$row['entity'];
        }
    }

    /* get all field-ids (.entity) and CSV-lists of field (.value) of all fields. this query could be more specific */
    $sql = "SELECT `CRMattributes`.`entity` FROM `CRMattributes` WHERE `CRMattributes`.`attribute` ='RepopulateByAjax' AND `CRMattributes`.`identifier` ='extrafield' AND `CRMattributes`.`value`='Yes'";
    $lor = db_GetArray($sql);

    /* loop through the matched fields - they need to be refreshed always regardless of what happened */
    foreach ($lor as $row) {
        $dstablename = get_ds_tablename_for_field($row['entity']);

        /* check if the field is in "my" table. if so, add to result set */
        if ($dstablename == $table) {
            $result[] = $table.'-f'.$row['entity'];
        }
    }

    /* get all field-ids (.efid) of fields having a requirement condition depending on the value of the field which changed */
    $sql = "SELECT `CRMextrafieldrequiredconditions`.`efid` FROM `CRMextrafieldrequiredconditions` WHERE `CRMextrafieldrequiredconditions`.field='EFID".$field_id."'";
    $conditional_fields = db_GetFlatArray($sql);

    /* all results are a direct hit, add them to the result set */
    foreach ($conditional_fields as $cond) {
        $result[] = get_ds_tablename_for_field($field_id).'-f'.$cond;
    }

    /* slordig! soms komt de "eigen" fqn mee, dat gaat fout! */
    $endresult = array();
    foreach ($result as $field) {
        if ($field != $fqn) {
            $endresult[] = $field;
        }
    }

    return $result;
}
function ef_tabletype_to_ds_tablename($ef_tabletype)
{
    /*
        This functions converts a  table type used in the fields table to a datastructure tablename
    */
    if (ipi($ef_tabletype)) {
        return 't'.$ef_tabletype;
    } elseif ($ef_tabletype == 'user') {
        return 'users';
    } elseif ($ef_tabletype == 'group') {
        return 'groups';
    } elseif ($ef_tabletype == 'files') {
        return 'files';
    }
}
function ds_tablename_to_journal_tabletype($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table type used in the journal table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', 'flextable', $ds_tablename);
    } elseif ($ds_tablename == 'users') {
        return 'user';
    } elseif ($ds_tablename == 'groups') {
        return 'group';
    } else {
        return $ds_tablename;
    }
}
function ds_tablename_to_datajournal_tabletype($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table type used in the datajournal table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', '', $ds_tablename);
    } elseif ($ds_tablename == 'users') {
        return 'user';
    } elseif ($ds_tablename == 'groups') {
        return 'group';
    } else {
        return $ds_tablename;
    }
}
function ds_tablename_to_binfiles_table($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table used in the binfiles table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', 'flextable', $ds_tablename);
    } else {
        return $ds_tablename;
    }
}
function binfiles_table_to_ds_tablename($binfiles_tabletype)
{
    /*
        This functions converts a binfiles tabletype to the tablename used in the datasctructure
    */
    if (ipi(str_replace('flextable', '', $binfiles_tabletype))) {
        return str_replace('flextable', 't', $binfiles_tabletype);
    } else {
        return $binfiles_tabletype;
    }
}
function check_if_record_can_be_deleted($table, $recordid)
{
    /*
        This function checks if the user is allowed to change and therefore delete a record.
    */
    if (!$GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_delete']) {
        return 'This record cannot be deleted because you are not allowed to delete records in this table';
    } elseif ($recordid == '_new_') {
        return 'This record cannot be deleted because you cannot delete non-existing records';
    } elseif (CheckTableRecordAccess($table, $recordid) != 'ok') {
        return "This record cannot be deleted because access to this record didn't work out (non-existing record or table, or no access: ".CheckTableRecordAccess($table, $recordid).')';
    } elseif (CheckIfRecordHasChilds($table, $recordid)) {
        return 'This record cannot be deleted because there are still other records referring to it.';
    } else {
        /* nothing means OK */
        return '';
    }
}
function delete_record($table, $recordid, $skip_rules = false)
{
    /* No authentication. If not sure, don't use this function. */
    $sql = '';

    if ($GLOBALS['SystemDataStructure'][$table]['logical_delete'] == true) {

        /* this means the table uses the field "deleted" y|n to mark a record as deleted. */

        $sql = 'UPDATE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database']."`.`deleted` = 'y' ";

        $sql .= ' WHERE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'";

        /* we don't delete journal information so we even journal this event */
        DataJournal($recordid, '?', 'y', 'deleted', ds_tablename_to_datajournal_tabletype($table));

        log_msg('DELETE UNDO: UPDATE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database']."`.`deleted` = 'n' WHERE `CRM".$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'");
    } else {
        /* in this case, records are physically deleted from a table */

        $sql = 'DELETE FROM `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` WHERE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'";

        /* we don't delete journal information so we even journal this event */
        DataJournal($recordid, '?', 'y', 'physically-deleted', ds_tablename_to_datajournal_tabletype($table));
    }

    mcq($sql);

    /* delete files along with this record. files are always deleted physically */
    $sql = "DELETE `CRMbinfiles`.*, `CRMblobs`.* FROM `CRMbinfiles`, `CRMblobs` WHERE `CRMbinfiles`.`fileid` = `CRMblobs`.`fileid` AND `CRMbinfiles`.`koppelid` = '".$recordid."' AND `CRMbinfiles`.`type` = '".ds_tablename_to_binfiles_table($table)."'";

    /* put a message in the journal */
    journal($recordid, 'This record was deleted', ds_tablename_to_journal_tabletype($table));

    if (!$skip_rules) {
        /* add to-do's for rules */
        AddToDo($table.'-changed', '', $recordid, '');
        AddToDo($table.'-deleted', '', $recordid, '');
    }

    /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
    check_admin_changes_after_change($table);
}
function delete_record_including_childs($table, $records, $return_list_only = false, $skip_rules = false)
{
    /*
        This function deletes and record and all if its childs, but only if access is allowed. The $return_list_only
        variable can be used the return an array of what WOULD be done, including the result of the check per table.
        record.php uses this flag to display an overview to the user. This function is secured; it will not delete
        records the user doesn't have access to.
    */

    /* this became a constant - we're always descending to childs */
    if (!is_array($GLOBALS['SystemDataStructure'])) {
        $GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
    }
    $descend = true;

    $returntrue = false;

    /* $records should contain a simple non-associative array of record-id's to delete */
    if (is_array($records)) {

        /* some initial vairable values */
        $list = array();
        $list[$table]['ok'] = 0;
        $list[$table]['nok'] = 0;

        /* loop through records and check access */
        foreach ($records as $recordid) {

            /* access should be "ok" and general table-record-access should be allowed to be able to delete this record as were it childless */
            if (CheckTableRecordAccess($table, $recordid) == 'ok' && $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_delete'] == true) {

                /* access to THIS record is ok ignoring any childs */
                ++$list[$table]['ok'];

                if ($return_list_only) {
                    /* only return COUNT */
                        /* count and add the number of files associated with this record which will also be deleted should this record be deleted */

                        if ($GLOBALS['SystemDataStructure'][$table]['HasFileAttachments']) {
                            $sql = "SELECT COUNT(*) FROM CRMbinfiles WHERE type='".ds_tablename_to_binfiles_table($table)."' AND koppelid='".$recordid."'";

                            if (CheckTableRecordAccess($table, $recordid) == 'ok') {
                                $list[$table]['files'] += db_GetValue($sql);
                            } else {
                                ++$list[$table]['nok'];
                            }
                        }

                        /* check to see if the user is deleting a folder */
                        if ($table == 'files' && GetFileType($recordid) == '{{{folder}}}') {
                            /* the user is deleteing a folder (and nothing else) */
                            $sql = "SELECT fileid FROM CRMbinfiles WHERE folder='".$recordid."'";
                            foreach (db_GetFlatArray($sql) as $file) {
                                unset($file);
                                ++$list[$table]['ok'];
                            }
                        }
                } else {

                    /* actually delete the files here */

                    $fileids = db_GetFlatArray("SELECT fileid FROM CRMbinfiles WHERE type='".ds_tablename_to_binfiles_table($table)."' AND koppelid='".$recordid."'");

                    if (count($fileids) > 0) {
                        foreach ($fileids as $id) {
                            $skip_security = false;
                            $skip_revert = true;
                            DeleteFile($id, $skip_security, $skip_revert);
                        }
                    }
                }
            } else {

                /* access to this record is NOT ok. ignore files */
                ++$list[$table]['nok'];
            }
        }

        foreach ($records as $recordid) {
            if (CheckFunctionAccess('AllowedToUseDescendingDeletes') != 'ok' && CheckIfRecordHasChilds($table, $recordid) && !is_administrator()) {
                /* we're starting to descend here, but the user is not allowed to. quit. */
                $list[$table]['ok'] = 0;
                ++$list[$table]['nok'];

                return $list;
            }

            /* check if the user has access to delete THIS record regardless of the childs. ignore if a list must be returned */

            if (check_if_record_can_be_deleted($table, $recordid) != '' || $return_list_only) {

                //print '<br>check ' . $recordid . ' of table ' . $table . ' RLO: ' . $return_list_only . '<br>';

                if ($descend) {

                    /* get all fields referring to this field */
                    foreach ($GLOBALS['SystemDataStructure'][$table]['childs'] as $childtable => $ct) {
                        if ($childtable != $table && $childtable != 'journal' && $childtable != 'datajournal' && $childtable != 'users' && $childtable != 'groups') {
                            $fieldname = $GLOBALS['SystemDataStructure'][$childtable]['fields'][$ct[0]]['name_in_database'];

                            $deleted_part = '';
                            if ($GLOBALS['SystemDataStructure'][$childtable]['logical_delete'] == true) {
                                /* table supports logical deletes, include it in this query */
                                $deleted_part = ' `CRM'.$GLOBALS['SystemDataStructure'][$childtable]['name_in_database']."`.`deleted` = 'n' AND";
                            }

                            $sql = 'SELECT `CRM'.$GLOBALS['SystemDataStructure'][$childtable]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$childtable]['primary'].'` FROM `CRM'.$GLOBALS['SystemDataStructure'][$childtable]['name_in_database'].'` WHERE '.$deleted_part.' `CRM'.$GLOBALS['SystemDataStructure'][$childtable]['name_in_database'].'`.`'.$fieldname."`='".mres($recordid)."'";

                            //print "[" . $sql . "]";
                            $recordlist = db_GetFlatArray($sql);

                            foreach ($recordlist as $child_to_delete) {
                                if ($return_list_only) {
                                    foreach (delete_record_including_childs($childtable, array($child_to_delete), true) as $ret_table => $count) {
                                        $list[$ret_table]['ok'] += $count['ok'];
                                        $list[$ret_table]['nok'] += $count['nok'];
                                        $list[$ret_table]['files'] += $count['files'];
                                        //print "Add count " . $count . " to table $ret_table <br>";
                                    }
                                } else {
                                    /* actually call delete funtion */
                                    delete_record_including_childs($childtable, array($child_to_delete), false);
                                    //DA("Call elete_record_including_childs($childtable, array(" . string_r($child_to_delete) . ", false);");
                                }
                            }
                        } else {
                            // print "loop prevention";
                        }
                    }
                } // end if decend
            } else {
                // end check_if_record_can_be_deleted

                // denied. misschien loggen?
                //print "check_if_record_can_be_deleted($table, $recordid) returns ". check_if_record_can_be_deleted($table, $recordid) . "<br>";
            }

            if (!$return_list_only && $list[$table]['nok'] == 0) {

                /* check to see if the user is deleting a folder */
                if ($table == 'files' && GetFileType($recordid) == '{{{folder}}}') {
                    /* the user is deleteing a folder (and nothing else) */
                    $sql = "SELECT fileid FROM CRMbinfiles WHERE folder='".$recordid."'";
                    foreach (db_GetFlatArray($sql) as $file) {
                        delete_record($table, $file, $skip_rules);
                    }
                }

                /* The record we started with can be deleted */
                $returntrue = true;
                // print "Actually delete: $table-$recordid";
                if (check_if_record_can_be_deleted($table, $recordid) == '') {
                    delete_record($table, $recordid, $skip_rules);
                }
            } else {
                //	print "DENIED $table $recordid -- return_list_only: $return_list_only, returntrue: $returntrue, aantal nok: " . $list[$table]['nok'] . "<br>";
            }
        } // end foreach

        if ($returntrue) {
            return true;
        } elseif ($return_list_only) {
            return $list;
        } else {
            return false;
        }
    } else {
        log_msg('ERROR: delete_record_including_childs called with something else than an array of records');
    }
}
function check_admin_changes_after_change($table, $identifier = '')
{
    if (is_administrator()) {
        /* check mysql triggers */

        $tables = array('fields', 'tables');
        if (in_array($table, $tables)) {
            if (GetSetting('DevelopmentOrProduction') == 'Development') {
                SetAttribute('system', 'TriggerQueriesForDeploy', return_mysql_trigger_queries(true, false, true), 1, false);
            }
            require_once 'adminfunctions.php';
            $triggers = return_mysql_trigger_queries();
            foreach ($triggers as $sql) {
                mcq($sql);
            }
        }

        /* check datastructure drop */

        $tables = array('fields', 'tables', 'fieldconditions', 'fieldrequiredconditions', 'languages', 'settings', 'users', 'profiles', 'extendedprivsgroup', 'extendedprivsuser', 'attributes', 'tableaccessrules', 'templateaccessrules', 'moduleaccessrules', 'calendaraccessrules', 'fieldaccessrules', 'calendaraccessrules');

        if (in_array($table, $tables)) {
            switch ($table) {
            // dit staat nu uit, omdat er meerdere gegevens meegewogen moeten worden, dat lukt nu niet op een gestructureerde manier
            /*
            case "users":
                // drop DS only for this user
                DropPersonalDataStructure($identifier);
                break;
            case "profiles":
                // drop DS for all users in the same group
                foreach (GetUsersOfGroup(GetGroup($identifier)) AS $user_id) {
                    DropPersonalDataStructure($user_id);
                }
                break;
            case "attributes":
                // drop all DS when attibute identifier matches list
                $valid_identifiers = array("group", "extrafield", "flextable", "system");
                if (in_array($identifier, $valid_identifiers)) {
                    DropPersonalDataStructure();
                }
            */
            default:

                /* drop all DS */
                //AddToDo("DropAllDataStructures", "", $0, "");
    //			DropPersonalDataStructure();
                break;
            }
        }
    }
}
function add_default_values_for_fields($table, $recordid, $skip_fields)
{

    /*
        adds default values for fields just after creating a new record
    */

    /* use system data structure, we need ALL fields, not only visiable */
    CheckAndCreateSystemDataStructure();

    /* it's now in the databases */
    $sds = unserialize(db_GetValue("SELECT datastructure FROM CRMdatastructures WHERE userid=0 AND datastructure != ''"));

    $fields_to_update = array();
    $to_datajournal = array();
    $to_journal = '';

    /* loop through all fields in this table */
    foreach ($sds[$table]['fields'] as $dsname => $field) {
        if (!in_array($dsname, $skip_fields)) {
            if ($field['defaultvalue'] != '') {
                $val = get_parsed_template_with_given_template($table, $recordid, array(), '', false, '', $field['defaultvalue'], 'DEFAULTVALUE');

                if ($val != '') {
                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$dsname]['fieldtype_in_extrafields_table'] != 'date/time' && $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$dsname]['fieldtype_in_extrafields_table'] != 'date (iso)') {

                        /* validate the field value */
                        list($result, $message, $value) = ValidateFieldInputDataStructureOnly($table.'-'.$dsname, $val, true, $recordid);
                    } else {
                        $result = 'ok';
                        $value = $val;
                    }

                    /* everything is ok */
                    if ($result == 'ok') {
                        $fields_to_update[$dsname] = $value;

                        $to_datajournal[] = array($value, $dsname);
                        $to_journal .= 'Default value of '.$dsname.' ['.$sds[$table]['fields'][$dsname]['name'].'] set to [';

                        if ($sds[$table]['fields'][$dsname]['relation_to'] == 'users') {
                            $to_journal .= GetUserName($value);
                        } else {
                            $to_journal .= $value;
                        }

                        $to_journal .= "]\n";
                    } else {
                        $to_journal .= "Default value didn't validate: ".$value;
                    }
                }
            }
        } else {
            //		print "SKIP FIELD " . $dsname . " WAS ALREADY POSTED";
        }
    }

    if (count($fields_to_update) > 0) {
        $sql = 'UPDATE `CRM'.$sds[$table]['name_in_database'].'` SET ';
        $sqlparts = array();

        foreach ($fields_to_update as $field => $value) {
            $sqlparts[] = '`CRM'.$sds[$table]['name_in_database'].'`.`'.$sds[$table]['fields'][$field]['name_in_database']."` = '".mres($value)."' ";
        }

        $sql .= implode(',', $sqlparts);

        $sql .= ' WHERE '.$sds[$table]['primary'].' = '.$recordid;

        mcq($sql);

        if ($to_journal != '') {
            journal($recordid, $to_journal, ds_tablename_to_journal_tabletype($table));
        }
        foreach ($to_datajournal as $dj) {
            DataJournal($recordid, '{{def}}',  $dj[0], $dj[1], ds_tablename_to_datajournal_tabletype($table));
        }

//		print '<h1>' . $sql . '</h1>';
//		DA($skip_fields);
        //exit;
    }
}
function run_isolated_php_code($code, $table = '', $record = '', $pass = array())
{
    global $lang;

    foreach ($pass as $varname => $varvalue) {
        $$varname = $varvalue;
    }

    if (ipi($recordid) && is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) {

        /* parse the code */
        $code = unescape_interleave_characters(ParseTemplateRecordData($code, $table, $record));

        /* compatibility */
        /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */

        $recordid = $record;
        $eid = $record;
        $flextablerecord = $record;

        if (ipi(str_replace('f', '', $table))) {
            $flextable = str_replace('f', '', $table);
            $flextableid = str_replace('f', '', $table);
        }
    }

    /* execute the code */
    /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
    $result = false;

    @eval($code);

    return $result;
}

function validate_field_input($fqn, $value, $record, $check_for_required_fields = false)
{
    global $lang;

    if ($GLOBALS['DEVELOPMENT']) {

        /* to test only server-side validation */
        if ($_POST['form_submitted'] != 1) {
            $return = array();
            $return['result'] = 'ok';
            $return['message'] = ' niet gechecked dude ';
            $return['corrected_value'] = $value;

            return array($return['result'], $return['message'], $return['corrected_value']);
        }
    }
    /*
        The general field value validator

    */

    /* explode fqn */
    list($table, $field) = explode('-', $fqn);

    /* we ALWAYS trim leading and trailing spaces from values */
    if (!is_array($value)) {
        $value = trim($value);
    }

    /* copy (by ref) the interesting part of the data structure */
    $prop = &$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field];

    /***************************************************************************************/
    /*
     * 					LEGACY, TODO/FIX/REMOVE
     */

    /* convert dates and date/time to ISO, so we can change the callers later on as soon as ValidateFieldInput is eliminated */
    if ($prop['fieldtype'] == 'date') {

        /* dates are given as dd-mm-yyyy */
        if ($prop['dateformat_in_database'] == 'yyyy-mm-dd' || $prop['dateformat_in_database'] == 'dd-mm-yyyy') {
            $value = NLDate2INTLDate($value);
        }
    }
    /* BELOW THIS LINE ALL DATES ARE TREATED AS ISO */
    /***************************************************************************************/

    /* set initial response */
    $return = array('result' => 'undetermined', 'message' => $fqn.' failed validation type:'.$prop['fieldtype'].' name: '.$prop['name'], 'corrected_value' => $value);

    /* we don't validate buttons in this function */
    /* why not? TODO */
    if ($prop['formelementtype'] == 'button') {
        $return = array('result' => 'ok', 'message' => '', 'corrected_value' => '');
    } else {

        /* check to see if there is a custom validation php set */
        if ($prop['CustomValidationFunctionPHP'] != '') {

            /* run code isolated / in own scope to prevent other vars being changed */
            $result = run_isolated_php_code($prop['CustomValidationFunctionPHP'], $table, $record, array('value' => $value));

            /* this is nasty. the code can result in two types;

                $result as array having the same structure as in this function
                $result as string simply  containing "ok" or "nok
            */

            if (is_array($result)) {
                if ($result[0] == 'ok' || $result[0] == 'nok') {
                    $return['result'] = $result[0];
                    $return['message'] = $result[1];
                    if (isset($result[2])) {
                        $return['corrected_value'] = $result[2];
                    }
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Custom validation function returned invalid response: '.string_r($result);
                }
            } else {
                if ($result == 'ok' || $result == 'nok') {
                    $return['result'] = $result[0];
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Custom validation function returned invalid response: '.$result;
                }
            }
        }

        /* determine if the submitted value is empty */
        $is_empty = true;

        /* $value is already trimmed */
        if (!is_array($value) && $value !== '') {

            /* plain values */
            $is_empty = false;
        } elseif (is_array($value)) {

            /* multiselect values */
            foreach ($value as $val) {

                /* we need to trim $val here */
                if (trim($val) !== '') {
                    $is_empty = false;
                }
            }
        }

        /* now, check requirement. if a field is required, but not entered, nothing else matters */
        if ($return['result'] == 'undetermined' && $check_for_required_fields && $is_empty && $prop['required'] && $prop['formelementtype'] != 'checkbox') {
            if (count($prop['requiredconditions']) == 0) {

                /* no conditions on requirement; it just cannot be empty */
                $return['result'] = 'nok';
                //$return['message']  = $lang['requiredfields'];
            } elseif (ipi($record)) {

                /* the field is required, but only in certain conditions */

                /*
                 *
                 *
                 * VAN DIT STUKJE KLOPT NIKS!!
                 *
                 *
                 *   [requiredconditions] => Array
                                    (
                                        [0] => Array
                                            (
                                                [otherfield] => t78-f1375
                                                [value] =>
                                                [mustbetrue] =>
                                            )

                                    )
                    */

                $fields_to_fetch = array();
                foreach ($prop['requiredconditions'] as $rc) {
                    $fields_to_fetch[] = $rc['otherfield'];
                }
                $selectallfields = false;
                $bypass_record_access_by_sql = false;
                $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

                $fv = false;
                if (isset($_REQUEST[$rc['otherfield'].'-'.$record])) {
                    $fv = $_REQUEST[$rc['otherfield'].'-'.$record];
                }
                /* DIRTY HACK */
                /*
                        zie ook functions-ajax.php.
                */
                if (isset($GLOBALS['submitted_values'][$rc['otherfield']])) {
                    $fv = $GLOBALS['submitted_values'][$rc['otherfield']];
                }

                foreach ($prop['requiredconditions'] as $rc) {
                    if ($fv !== false) {
                        $current_record[$rc['otherfield']] = $fv;
                    }
                    if (
                        ($rc['mustbetrue'] && $current_record[$rc['otherfield']] != $rc['value'])
                        ||
                        (!$rc['mustbetrue'] && $current_record[$rc['otherfield']] == $rc['value'])
                        ) {
                        $return['result'] = 'nok';
                        //	$return['message']  = $lang['requiredfields'];
                    }
                }
            } else {
                /* we don't support requirement conditions on new records. do nothing. */
            }
        }

        if ($return['result'] == 'undetermined' && $is_empty && !$prop['required']) {

            /* the value is empty, and it's not required in any way, and CustomValidationFunctionPHP didn't NOK it. we're fine */
            $return['result'] = 'ok';
            $return['message'] = '';

        /* at this point we know it's a normally entered field. earlier checks on CustomValidationFunctionPHP and
           requirement didn't result in a NOK, and it's not empty */
        } elseif ($return['result'] == 'undetermined') {

            /* first, check if the field has options set in the datastructure. if so, use them as final check and ignore anything else */
            if (is_array($prop['options']) && count($prop['options']) > 0) {
                $some_failed = false;
                $failed = array();

                /* this catches all fields who have options set */

                if ($prop['multiselect']) {
                    foreach ($value as $val) {
                        if (!in_array($val, $prop['options']) && $val != '') {
                            $some_failed = true;
                            $failed[] = $val;
                        }
                    }
                } else {
                    if (!in_array($value, $prop['options'])) {
                        $some_failed = true;
                        $failed[] = $value;
                    }
                }
                if ($some_failed) {
                    $return['result'] = 'nok';
                    $return['message'] = 'Given value '.implode(',', $failed).' not found in options';
                } else {
                    $return['result'] = 'ok';
                    $return['message'] = '';
                    //$return["message"] = $field . ": " . $prop['fieldtype_in_extrafields_table'] . ": approved";
                }
            } elseif ($prop['multiselect'] && !$prop['relation_to']) {

                /*
                  this MUST be either "Drop-down (populate by code multiselect)" or "SQL Query (multiselect)".
                  we do not validate these types

                */
                $return['result'] = 'ok';
                $return['message'] = '';
                //$return["message"] = $field . ": " . $prop['fieldtype_in_extrafields_table'] . ": ignored";
            } elseif ($prop['relation_to'] != '') {

                /* this catches all fields referring to another table: ref-to-flex, user, group */
                $check_if_exists_recordids = false;

                if (!$prop['multiselect']) {

                    /* select_refer_field_options always needs an array with selected values */
                    $check_if_exists_recordids = array($value);
                } else {

                    /* multiselect values are given as an array */
                    $check_if_exists_recordids = $value;
                }

                /* check if it's an array in case of multiselect. single select will always be an array, see above */
                if (!is_array($check_if_exists_recordids)) {

                    /* not an array, but it should be. will have something to do with a multiselect not being passed as an array */
                    log_msg('ERROR: Field '.$field.' is a multiselect but the value given to validate_field_input is not an array');
                    $return['result'] = 'nok';
                    $return['message'] = "Given value should be an array, but it isn't";
                } else {

                    /* get possible values. select_refer_field_options will add the currently selected value(s) if they are not in the result set */
                    $possible_values = select_refer_field_options($table, $field, $record, false, $check_if_exists_recordids);

                    /* cycle trough options and check if all selected values are in the result set */
                    $found = true;
                    $failed = array();
                    foreach ($check_if_exists_recordids as $check_if_exists_recordid) {
                        if (ipi($check_if_exists_recordid) && !array_key_exists($check_if_exists_recordid, $possible_values)) {

                            /* given value not found; the user is not allowed to select this value; deny */
                            $found = false;
                            $failed[] = $check_if_exists_recordid;
                        }
                    }

                    /* if all went well... */
                    if ($found) {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                    } else {

                        /* no message, defaults to default message */
                        //$return["message"] = "One of the given values not found in result of select_refer_field_options.";

                        $return['result'] = 'nok';
                        $return['message'] = '';

                        /* log this, it shoudn't happen */
                        log_msg('ERROR: A user selected a value in a reference field, the value is not in the list of options.');
                    }
                }
            } elseif ($prop['fieldtype'] == 'text') {

                /*

                these DS text-types are not validated but treated as text:

                fieldtype_in_extrafields_table == 'Drop-down (populate by code)'
                fieldtype_in_extrafields_table == 'SQL Query'

                 */

                if (ipi($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumLength']) && strlen($value) > $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumLength']) {

                    /* the length of the given value exceedes the maximum length of the field */
                    $return['result'] = 'nok';

                    /* no message, defaults to default message */
                    $return['message'] = '';
                } else {

                    /* hyperliks */
                    if ($prop['fieldsubtype'] == 'link') {
                        if (substr($value, 0, 7) != 'http://' && substr($value, 0, 8) != 'https://') {
                            $value = 'http://'.$value;
                        }

                        if (IsValidURL($value)) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $value;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* e-mail addresses */
                    } elseif ($prop['fieldsubtype'] == 'mail') {
                        if (ValidateEmail($value)) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $value;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* telephone numbers */
                    } elseif ($prop['fieldsubtype'] == 'tel') {
                        list($ok, $ret) = IsValidPhonenumber($value);

                        if ($ok) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $ret;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* nothing more to check */
                    } else {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                    }
                }
            } elseif ($prop['fieldtype'] == 'number') {
                if (is_numeric($value)) {

                    /* check for minimum and maximum value */
                    if (
                        /* maximum */
                        (is_numeric($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumValue']) && $value > $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumValue'])
                        ||
                        /* minimum */
                        (is_numeric($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MinimumValue']) && $value < $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MinimumValue'])
                       ) {
                        $return['result'] = 'nok';
                        /* no message, defaults to default message */
                        $return['message'] = '';

                        /* we use HARD CODED decimal char "." here. IS THIS CORRECT? */
                        $return['corrected_value'] = number_format($value, $prop['decimals'], '.', '');
                    } else {
                        $return['result'] = 'ok';
                        $return['message'] = '';

                        /* we use HARD CODED decimal char "." here. IS THIS CORRECT? */
                        $return['corrected_value'] = number_format($value, $prop['decimals'], '.', '');
                    }
                } else {

                    /* not even numeric */
                    $return['result'] = 'nok';

                    /* no message, defaults to default message */
                    $return['message'] = '';
                }
            } elseif ($prop['fieldtype'] == 'date') {

                /* check if it's date or date-time. yyyy-mm-dd and dd-mm-yyyy mean date */
                if ($prop['dateformat_in_database'] == 'yyyy-mm-dd' || $prop['dateformat_in_database'] == 'dd-mm-yyyy') {
                    if (is_valid_iso_date($value)) {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                        $return['corrected_value'] = from_iso_datetime($value);
                    } else {
                        $return['result'] = 'nok';
                        $return['message'] = '';
                    }

                /* the only option which can refer to date/time */
                } elseif ($prop['dateformat_in_database'] == 'yyyy-mm-dd hh:mm:ss') {

                    /* a date-and-time field */
                    $test = StrDateTime2IsoDateTime($value);

                    if ($test != '') {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                        //$return["message"] = 'OK waarde: ' . $value . ' result: ' . $test;
                        $return['corrected_value'] = trim($test);
                    } else {
                        $return['result'] = 'nok';
                        /* no message, defaults to default message */
                        $return['message'] = '';
                    }
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Unknown format ('.$prop['dateformat_in_database'].')';
                }
            }
        }
    }

    //$return["message"] .= "run end req:" . $check_for_required_fields;
    if ($return['result'] == 'undetermined') {
        $return['result'] = 'nok';
        $return['message'] .= ' (UNDETERMINED)';
    }
    if ($return['result'] == 'nok') { // && $GLOBALS['SesMem']['debug']) {
        $return['message'] .= '	FIELD '.$field.' TABLE '.$table.' VALUE '.$value.' TYPE '.$prop['fieldtype'];
    }

    return array($return['result'], $return['message'], $return['corrected_value']);
}

function generate_hideandshow_javascript($sourcedstable)
{
    /* TODO deze functie hobbelt nu 3x door de lijst met rules heen. dat kan makkelijk naar 1x */

    /* this variable is not used - in the future we should (maybe) only print the appropriate fields - for now, we do all */
    $sourcedstable = false;

    /* now we will loop through all visual_only extended privileges */
    $sqladd = '';
    $functions_created = array();

    /* add group number if the group id > 0 */
    if (ipi($GLOBALS['GROUPID'])) {
        $sqladd .= ','.$GLOBALS['GROUPID'];
    }

    /* admin rules (groupid=0) are appliccable to everybody */
    $sql = "SELECT * FROM `CRMextendedprivsgroup` WHERE `CRMextendedprivsgroup`.`visual_only` = 'y' AND `CRMextendedprivsgroup`.`only_apply_for_field` > 0 AND `CRMextendedprivsgroup`.`groupid` IN (0".$sqladd.')';
    $list = db_GetArray($sql);
    $rules = array();
    $todo = array();

    /* flatten this array */
    foreach ($list as $row) {
        $fqn = get_ds_tablename_for_field($row['only_apply_for_field']).'-f'.$row['only_apply_for_field'];

        /* declare array if not already so */
        if (!is_array($rules[$fqn])) {
            $rules[$fqn] = array();
        }
        if (!is_array($rules[$fqn][$row['rule']])) {
            $rules[$fqn][$row['rule']] = array();
        }

        $forfqn = get_ds_tablename_for_field($row['field']).'-f'.$row['field'];

        /* group by field first, type of rule (invisible or readonly) next */
        $rules[$fqn][$row['rule']][] = array('field' => $forfqn, 'operand' => $row['operand'], 'value' => $row['value'], 'rule' => $row['rule']);
        if (!in_array($forfqn, $todo)) {
            $todo[] = $forfqn;
        }
    }

    /* compose javascript */
    $js = '';
    $output = '';
    $extrajs = '';
    $js .= "

function hideshowallelements() {
	/* rules for invisibility - THIS IS NOT SECURITY */

	var currenttable = $('#JS_concerningtable').val();
	var currentrecord = $('#JS_concerningrecord').val();

	var made_readonly = [];

	if (currenttable != '' && currentrecord != '') {

";

    /* in this loop we create some js which will fetch the value of the fields, but only if we're looking
       at fields in the same table as the form is for. we only know this client-side. */
    $js .= "		/* get the field values, as long as the field is in the same table as the form */\n";
    foreach ($todo as $field) {
        $done[] = $field;

        list($dstable, $dsfield) = explode('-', $field);

        $js_element_to_check = str_replace('-', '_', $field);

        $js .= "		if (currenttable == '".$dstable."' && typeof ".$js_element_to_check." == 'undefined') { var ".$js_element_to_check." = find_field_value('".$field."', currentrecord); }\n";
    }
    //$js .= string_r($rules);
    /* rule example

        [t4-f182] => Array
            (
                [invisible] => Array
                    (
                        [0] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Wacht op dagboekbericht
                                [rule]  =>  invisible
                            )

                        [1] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Dagboekbericht geplaatst
                                [rule]  =>  invisible
                            )

                        [2] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Afgerond
                                [rule]  =>  invisible
                            )

                        [3] => Array
                            (
                                [field]  =>  t4-f76
                                [operand]  =>  IS
                                [value]  =>  geen
                                [rule]  =>  invisible
                            )

                    )

            )
    */
    $megaif = array();
    foreach ($rules as $field => $rule) {
        $megaif[$field]['invisible'] = array('is' => array(), 'is not' => array());
        $megaif[$field]['readonly'] = array('is' => array(), 'is not' => array());

        foreach ($rule['invisible'] as $eval) {
            if ($eval['operand'] == 'IS NOT') {
                $megaif[$field]['invisible']['is not'][$eval['field']][] = inline_jsencode($eval['value']);
            } elseif ($eval['operand'] == 'IS') {
                $megaif[$field]['invisible']['is'][$eval['field']][] = inline_jsencode($eval['value']);
            }
        }
        foreach ($rule['readonly'] as $eval) {
            if ($eval['operand'] == 'IS NOT') {
                $megaif[$field]['readonly']['is not'][$eval['field']][] = inline_jsencode($eval['value']);
            } elseif ($eval['operand'] == 'IS') {
                $megaif[$field]['readonly']['is'][$eval['field']][] = inline_jsencode($eval['value']);
            }
        }
    }
//	$js .= string_r($megaif);
    /* megaif sample

    [t21-f305] => Array
           (
               [invisible] => Array
                   (
                       [is]  =>
                       [is not]  =>
                   )

               [readonly] => Array
                   (
                       [is]  =>
                       [is not] => Array
                           (
                               [t21-f395] => Array
                                   (
                                       [0]  =>  Yes
                                   )

                               [t21-f318] => Array
                                   (
                                       [0]  =>  1
                                   )

                               [t21-f302] => Array
                                   (
                                       [0]  =>  RORO
                                   )

                           )

                   )

           )

      */

    foreach ($megaif as $fqn => $data) {
        $k = false;
        $f = false;
        if (count($data['readonly']['is not']) > 0 || count($data['readonly']['is']) > 0) {
            $js .= '
				if (';
            if (count($data['readonly']['is not']) > 0) {
                $js .= '(';
                $k = false;
                foreach ($data['readonly']['is not'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;
                    $js .= ' (';
                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' && ';
                        }
                        $f = true;
                        $js .= ' '.str_replace('-', '_', $me)." != '".$value."'";
                    }
                    $js .= ')';
                }
                $js .= ')';
            }

            if (count($data['readonly']['is']) > 0) {
                if ($k) {
                    $js .= '|| ';
                }

                $k = false;
                foreach ($data['readonly']['is'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;

                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' || ';
                        }
                        $f = true;
                        $js .= ' '.str_replace('-', '_', $me)." == '".$value."'";
                    }
                }
            }

            $js .= ')';

            $js .= " {
	  				$('#ffc-" .$fqn." :input').prop(\"disabled\", true);
	  	        	$('#ffc-" .$fqn."').show(250);
					made_readonly.push('" .$fqn."');
	  			} else {
					$('#ffc-" .$fqn." :input').prop(\"disabled\", false);
	  	        	$('#ffc-" .$fqn."').show(250);
				}";
        }
    }
    foreach ($megaif as $fqn => $data) {
        $k = false;
        $f = false;
        if (count($data['invisible']['is not']) > 0 || count($data['invisible']['is']) > 0) {
            $js .= ' if (';

            if (count($data['invisible']['is not']) > 0) {
                $js .= '(';
                $k = false;
                foreach ($data['invisible']['is not'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;
                    $js .= ' (';
                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' && ';
                        }
                        $f = true;
                        $js .= ' '.str_replace('-', '_', $me)." != '".$value."'";
                    }
                    $js .= ')';
                }
                $js .= ')';
            }

            if (count($data['invisible']['is']) > 0) {
                if ($k) {
                    $js .= ' || ';
                }
                $k = false;

                foreach ($data['invisible']['is'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;

                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' || ';
                        }
                        $f = true;
                        $js .= ' '.str_replace('-', '_', $me)." == '".$value."'";
                    }
                }
            }
            $js .= ')';

            $js .= " {
				$('#ffc-" .$fqn." :input').prop(\"disabled\", true);
	        	$('#ffc-" .$fqn."').hide();
			} else {
				if ($.inArray('" .$fqn."', made_readonly) == -1) {
					$('#ffc-" .$fqn." :input').prop(\"disabled\", false);
				}
				$('#ffc-" .$fqn."').show(250);

			}";
        }
    }

    $js .= '

	}
}
$(document).ready(function() { hideshowallelements(); });
';
    //return "\n" . str_replace("\n", "", str_replace("\t", "", $js)) . "\n";
    return $js;
}
